import skimage as sk
import sklearn as skl
import numpy as np
import pandas as pd
import pickle as pkl

import copy
import h5py

from sklearn import cluster
from scipy.special import logsumexp
from matplotlib import pyplot as plt
from matplotlib import cm
from pulp import *
from ipywidgets import interactive, fixed, FloatSlider, IntSlider, IntRangeSlider, SelectMultiple

from .utils import pandas_hdf5_handler,writedir
from .trcluster import dask_controller

def get_labeled_data(kymo_arr,orientation):
    labeled_data = []
    for t in range(kymo_arr.shape[0]):
        if orientation == 1:
            flipped_arr = kymo_arr[t,::-1]
            labeled = sk.morphology.label(flipped_arr,connectivity=1)
        else:
            labeled = sk.morphology.label(kymo_arr[t],connectivity=1)
        unique_idx = np.unique(labeled[-1])
        remove_indices = unique_idx[unique_idx!=0]
        for remove_idx in remove_indices:
            labeled[labeled==remove_idx] = 0
        labeled_data.append(labeled)
    labeled_data = np.array(labeled_data)
    return labeled_data

def get_segment_props(labeled_data,interpolate_empty_trenches=False):
    all_centroids = []
    all_sizes = []
    all_bboxes = []
    
    for t in range(labeled_data.shape[0]):
        rps = sk.measure.regionprops(labeled_data[t])
        centroids = [rp.centroid for rp in rps]
        sizes = np.array([rp.area for rp in rps])
        bboxes = np.array([rp.bbox for rp in rps])
        centroids = np.array([[centroid[1],centroid[0]] for centroid in centroids])
        all_centroids.append(centroids)
        all_sizes.append(sizes)
        all_bboxes.append(bboxes)
    
    if interpolate_empty_trenches:
        empty_trenches = []
        for t in range(labeled_data.shape[0]):
            if len(all_sizes[t]) == 0:
                all_centroids[t] = all_centroids[t-1]
                all_sizes[t] = all_sizes[t-1]
                all_bboxes[t] = all_bboxes[t-1]
                empty_trenches.append(t)
        return all_centroids,all_sizes,all_bboxes,empty_trenches
    else:
        return all_centroids,all_sizes,all_bboxes
    
class scorefn:
    def __init__(self,headpath,segfolder,u_pos=0.2,sig_pos=0.4,u_size=0.,sig_size=0.2,w_merge=0.8):
        self.headpath = headpath
        self.segpath = headpath + "/" + segfolder
        self.u_pos,self.sig_pos = (u_pos,sig_pos)
        self.u_size,self.sig_size = (u_size,sig_size)
        self.w_merge = w_merge
        self.viewpadding = 0
    
    def fpos(self,xT,xt,bbox_T,bbox_t,u_pos=0.2,sig_pos=0.4):
        min_edge_T,min_edge_t = (np.min(bbox_T[:,0]),np.min(bbox_t[:,0]))
        l_T = xT[:,1]-min_edge_T
        l_t = xt[:,1]-min_edge_t
        del_l = np.subtract.outer(l_T,l_t)
        norm_del_l = del_l/l_t
        f_pos = -((norm_del_l-u_pos)**2)/(2*(sig_pos**2))
        return f_pos,norm_del_l

    def del_pos(self,centroids,bboxes):
        posscores = []
        for t in range(len(centroids)-1):
            posscore,_ = self.fpos(centroids[t+1],centroids[t],bboxes[t+1],bboxes[t],u_pos=self.u_pos,sig_pos=self.sig_pos)
            posscores.append(posscore.T)
        posscores = np.array(posscores)
        return posscores
        
    def del_split_pos(self,centroids,union_centroids,bboxes):
        posscores = []
        for t in range(len(centroids)-1):
            if len(union_centroids[t+1]) > 0:
                posscore,_ = self.fpos(union_centroids[t+1],centroids[t],bboxes[t+1],bboxes[t],u_pos=self.u_pos,sig_pos=self.sig_pos)
            else:
                posscore = np.array([[]])
            posscores.append(posscore.T)   
        posscores = np.array(posscores)
        return posscores
    
    def del_merge_pos(self,centroids,union_centroids,bboxes):
        posscores = []
        for t in range(len(centroids)-1):
            if len(union_centroids[t]) > 0:
                posscore,_ = self.fpos(centroids[t+1],union_centroids[t],bboxes[t+1],bboxes[t],u_pos=self.u_pos,sig_pos=self.sig_pos)
            else:
                posscore = np.array([[]])
            posscores.append(posscore.T)   
        posscores = np.array(posscores)
        return posscores
    
    def fsize(self,sizeT,sizet,u_size=0.,sig_size=0.2):
        sizediff = np.subtract.outer(sizeT,sizet)
        normsizediff = sizediff/sizet
        f_size = -((normsizediff-u_size)**2)/(2*(sig_size**2))
        f_size[normsizediff<-0.1] = -np.inf

        return f_size,normsizediff
        
    def del_size(self,sizes):
        sizescores = []
        for t in range(len(sizes)-1):
            sizescore,_ = self.fsize(sizes[t+1],sizes[t],u_size=self.u_size,sig_size=self.sig_size)
            sizescores.append(sizescore.T)   
        sizescores = np.array(sizescores)
        return sizescores
    
    def del_half_size(self,sizes):
        sizescores = []
        for t in range(len(sizes)-1):
            size_ti = sizes[t]/2.
            size_tf = sizes[t+1]
            sizescore,_ = self.fsize(size_tf,size_ti,u_size=self.u_size,sig_size=self.sig_size)
            sizescores.append(sizescore.T)   
        sizescores = np.array(sizescores)
        return sizescores

    def del_split_size(self,sizes,union_sizes):
        sizescores = []
        for t in range(len(sizes)-1):
            if len(union_sizes[t+1]) > 0:
                sizescore,_ = self.fsize(union_sizes[t+1],sizes[t],u_size=self.u_size,sig_size=self.sig_size)
            else:
                sizescore = np.array([[]])
            sizescores.append(sizescore.T)   
        sizescores = np.array(sizescores)
        return sizescores
    
    def del_merge_size(self,sizes,union_sizes):
        sizescores = []
        for t in range(len(sizes)-1):
            if len(union_sizes[t]) > 0:
                sizescore,_ = self.fsize(sizes[t+1],union_sizes[t],u_size=self.u_size,sig_size=self.sig_size)
            else:
                sizescore = np.array([[]])
            sizescores.append(sizescore.T)   
        sizescores = np.array(sizescores)
        return sizescores
    
    def get_unions(self,centroids,sizes):
        union_sizes = []
        union_centroids = []
        for t in range(len(sizes)):
            union_sizes_t = []
            union_centroids_t = []
            for k in range(1,sizes[t].shape[0]):
                union_size = sizes[t][k-1] + sizes[t][k]
                union_centroid = (sizes[t][k-1]*centroids[t][k-1] + sizes[t][k]*centroids[t][k])/union_size
                union_sizes_t.append(union_size)
                union_centroids_t.append(union_centroid)
            union_sizes_t = np.array(union_sizes_t)
            union_centroids_t = np.array(union_centroids_t)
            union_sizes.append(union_sizes_t)
            union_centroids.append(union_centroids_t)
        size_diffs = [size[1:]-size[:-1] for size in sizes]
        return union_centroids,union_sizes,size_diffs
        
    
    def compute_score_arrays(self,centroids,sizes,bboxes):
        union_centroids,union_sizes,size_diffs = self.get_unions(centroids,sizes)
        
        sizeij = self.del_size(sizes)
        posij = self.del_pos(centroids,bboxes)
        
        posik = self.del_split_pos(centroids,union_centroids,bboxes)
        sizeik = self.del_split_size(sizes,union_sizes)
        
        poski = self.del_merge_pos(centroids,union_centroids,bboxes)
        sizeki = self.del_merge_size(sizes,union_sizes)
        
        half_sizeij = self.del_half_size(sizes)
        
        pos_e = [item[:,-1:] for item in posij]
        size_e = [item[:,-1:] for item in half_sizeij]
        
        
        Cij = []
        Cik = []
        Cki = []
        Ce = []
        
        for t in range(len(posij)):
            
            f_ij = posij[t]+sizeij[t]
            f_ik = posik[t]+sizeik[t]
            f_ki = poski[t]+sizeki[t]
            f_e = pos_e[t]+size_e[t]
            
            pad_f_ki = np.pad(f_ki,((1,1),(0,0)),"constant",constant_values=0.)

            if f_ik.shape[0] == 0:
                f_ik = np.zeros(f_ij.shape,dtype=float)
                
            
            log_sum = np.concatenate((f_ij,f_ik,pad_f_ki[1:],pad_f_ki[:-1]),axis=1)

            log_sum[:,0:1] += f_e
                    
            log_sum = logsumexp(log_sum,axis=1,keepdims=True)
            undefined_logsum = (log_sum==-np.inf)[:,0]

            f_ij[undefined_logsum] = 0.
            f_ik[undefined_logsum] = 0.
            f_ki[undefined_logsum[:-1]] = 0.
            f_e[undefined_logsum] = 0.
                        
            log_prob_ij = f_ij-log_sum
            log_prob_ik = f_ik-log_sum
            log_prob_ki = f_ki-log_sum[:-1]
            log_prob_e = f_e-log_sum
            
            ij_score = (-log_prob_ij)-100.
            ij_score[ij_score>0] = 0.
            ij_score[ij_score<-100.] = -100.
            Cij.append(ij_score)
        
            ik_score = ((-log_prob_ik)-100.)
            ik_score[ik_score>0] = 0.
            ik_score[ik_score<-100.] = -100.
            Cik.append(ik_score)    
        
            ki_score = (1.+self.w_merge)*(-log_prob_ki-100.)
            ki_score[ki_score>0] = 0.
            ki_score[ki_score<-100.] = -100.
            Cki.append(ki_score)
        
            e_score = ((-log_prob_e)-100.)
            e_score[e_score>0.] = 0.
            e_score[e_score<-100.] = -100.
            Ce.append(e_score)
                                    
        return Cij,Cik,Cki,Ce,posij,posik,poski
    
    def compute_neighbor_stats(self,centroids,bboxes,sizes):
        neighbor_dists = []
        neighbor_sizes = []
        for t in range(len(centroids)-1):
            _,norm_del_l = self.fpos(centroids[t+1],centroids[t],bboxes[t+1],bboxes[t],u_pos=self.u_pos,sig_pos=self.sig_pos)
            _,normsizediff = self.fsize(sizes[t+1],sizes[t],u_size=self.u_size,sig_size=self.sig_size)
            
            pos_norm_del_l = copy.copy(norm_del_l)
            pos_norm_del_l[pos_norm_del_l<0.] = np.inf
            
            pos_norm_sizediff = copy.copy(normsizediff)
            pos_norm_sizediff[pos_norm_sizediff<0.] = np.inf
            
            neighbor_idx = np.argmin(pos_norm_del_l,axis=0)

            neighbor_dist = pos_norm_del_l[neighbor_idx,np.arange(pos_norm_del_l.shape[1])].flatten()
            neighbor_size = pos_norm_sizediff[neighbor_idx,np.arange(pos_norm_sizediff.shape[1])].flatten()
            
            neighbor_dist = neighbor_dist[neighbor_dist != np.inf].tolist()
            neighbor_size = neighbor_size[neighbor_size != np.inf].tolist()
            
            neighbor_dists += neighbor_dist
            neighbor_sizes += neighbor_size
        neighbor_dists = np.array(neighbor_dists)
        neighbor_sizes = np.array(neighbor_sizes)
        return neighbor_dists,neighbor_sizes
    
    def plot_neighbor_stats(self,ax_dist,ax_size,centroids,bboxes,sizes):
        neighbor_dists,neighbor_sizes = self.compute_neighbor_stats(centroids,bboxes,sizes)
        
        dist_avg = np.mean(neighbor_dists)
        dist_std = np.std(neighbor_dists)
        size_avg = np.mean(neighbor_sizes)
        size_std = np.std(neighbor_sizes)
        
        ax_dist.set_xlim(0.,dist_avg+(3.*dist_std))
        ax_size.set_xlim(0.,size_avg+(3.*size_std))
        
        ax_dist.hist(neighbor_dists,bins=30)
        ax_dist.set_title('Nearest-neighbor Normalized Centroid Distance')
        ax_dist.axvline(x=self.u_pos,c="r",linewidth=3,zorder=10)
        ax_dist.axvspan(self.u_pos-self.sig_pos, self.u_pos+self.sig_pos, alpha=0.3, color='r',zorder=10)
        
        ax_size.hist(neighbor_sizes,bins=30)
        ax_size.set_title('Nearest-neighbor Normalized Size Difference')
        ax_size.axvline(x=self.u_size,c="r",linewidth=3,zorder=10)
        ax_size.axvspan(self.u_size-self.sig_size, self.u_size+self.sig_size, alpha=0.3, color='r',zorder=10)
        
        print("==========================================================")
        print("Mean Normalized Centroid Distance: " + str(dist_avg))
        print("Standard Deviation of Normalized Centroid Distance: " + str(dist_std))
        print("")
        print("Mean Normalized Size Difference: " + str(size_avg))
        print("Standard Deviation of Normalized Size Difference: " + str(size_std))
        
        plt.show()
    
    def plot_segmentation(self,ax,kymo_arr,t0=0,tf=-1,cmap='coolwarm'):
        img_shape = kymo_arr.shape[1:][::-1]
        x_dim = img_shape[0]+(self.viewpadding*2)
        ax.set_xlim(x_dim*t0,x_dim*tf)
        ax.set_ylim(0,img_shape[1])
        
        for t in range(t0,tf):
            padded_img = np.pad(kymo_arr[t][::-1],((0,0),(self.viewpadding,self.viewpadding)),'constant')
            ax.imshow(padded_img, extent=[t*x_dim, (t+1)*x_dim, 0, img_shape[1]])
    
    def plot_ij_scores(self,ax,kymo_arr,centroids,Cij,t0=0,tf=-1,cmap='coolwarm'):
        cmap_fn = cm.get_cmap(cmap, 100)
        img_shape = kymo_arr.shape[1:][::-1]
        x_dim = img_shape[0]+(self.viewpadding*2)
        if tf == -1:
            tf = len(centroids)-1
            
        ax.set_xlim(x_dim*t0,x_dim*tf)
        ax.set_ylim(0,img_shape[1])
        for t in range(t0,tf):
            padded_img = np.pad(kymo_arr[t][::-1],((0,0),(self.viewpadding,self.viewpadding)),'constant')
            ax.imshow(padded_img, extent=[t*x_dim, (t+1)*x_dim, 0, img_shape[1]])

            adjusted_centroids = centroids[t] + t*np.array([x_dim,0]) + np.array([self.viewpadding,0])
            ax.scatter(adjusted_centroids[:,0],adjusted_centroids[:,1],c="r",s=100,zorder=10)

            if t > t0:

                N_detect_i = len(centroids[t-1])
                N_detect_f = len(centroids[t])

                for i in range(N_detect_i):

                    for j in range(N_detect_f):

                        centroid_i = centroids[t-1][i]
                        centroid_f = centroids[t][j]

                        color_score = abs(Cij[t-1][i,j]/-100)

                        ax.plot([centroid_i[0] + (t-1)*x_dim + self.viewpadding, centroid_f[0] + t*x_dim + self.viewpadding],[centroid_i[1],centroid_f[1]],c=cmap_fn(color_score),linewidth=3)
        adjusted_centroids = centroids[tf] + tf*np.array([x_dim,0]) + np.array([self.viewpadding,0])
        ax.scatter(adjusted_centroids[:,0],adjusted_centroids[:,1],c="r")
        
    def plot_ik_scores(self,ax,kymo_arr,centroids,Cik,t0=0,tf=-1,cmap='coolwarm'):
        cmap_fn = cm.get_cmap(cmap, 100)
        img_shape = kymo_arr.shape[1:][::-1]
        x_dim = img_shape[0]+(self.viewpadding*2)
        if tf == -1:
            tf = len(centroids)-1        
        ax.set_xlim(x_dim*t0,x_dim*tf)
        ax.set_ylim(0,img_shape[1])
        for t in range(t0,tf):
            padded_img = np.pad(kymo_arr[t][::-1],((0,0),(self.viewpadding,self.viewpadding)),'constant')
            ax.imshow(padded_img, extent=[t*x_dim, (t+1)*x_dim, 0, img_shape[1]])

            merged_centroids = (centroids[t][:-1,:]+centroids[t][1:,:])/2

            adjusted_centroids = centroids[t] + t*np.array([x_dim,0]) + np.array([self.viewpadding,0])
            adjusted_merged_centroids = merged_centroids + t*np.array([x_dim,0]) + np.array([self.viewpadding,0])

            if t > t0:

                N_detect_i = len(centroids[t-1])
                N_detect_f = len(centroids[t])

                for i in range(N_detect_i):
                    for j in range(N_detect_f-1):

                        centroid_i = centroids[t-1][i]

                        centroid_k1 = centroids[t][j]
                        centroid_k2 = centroids[t][j+1]

                        color_score = abs(Cik[t-1][i,j]/-100)

                        ax.plot([centroid_i[0] + (t-1)*x_dim + self.viewpadding, centroid_k1[0] + t*x_dim + self.viewpadding],\
                                [centroid_i[1],(centroid_k1[1]+centroid_k2[1])/2],c=cmap_fn(color_score),linewidth=3)
            ax.scatter(adjusted_centroids[:,0],adjusted_centroids[:,1],c="r",s=100,zorder=10)
            ax.scatter(adjusted_merged_centroids[:,0],adjusted_merged_centroids[:,1],c="white",s=100,zorder=10)

        merged_centroids = (centroids[tf][:-1,:]+centroids[tf][1:,:])/2

        adjusted_centroids = centroids[tf] + tf*np.array([x_dim,0]) + np.array([self.viewpadding,0])
        adjusted_merged_centroids = merged_centroids + tf*np.array([x_dim,0]) + np.array([self.viewpadding,0])

        ax.scatter(adjusted_centroids[:,0],adjusted_centroids[:,1],c="r",s=100,zorder=10)
        ax.scatter(adjusted_merged_centroids[:,0],adjusted_merged_centroids[:,1],c="white",s=100,zorder=10)
    
    def plot_ki_scores(self,ax,kymo_arr,centroids,Cki,t0=0,tf=-1,cmap='coolwarm'):
        cmap_fn = cm.get_cmap(cmap, 100)
        img_shape = kymo_arr.shape[1:][::-1]
        x_dim = img_shape[0]+(self.viewpadding*2)
        if tf == -1:
            tf = len(centroids)-1        
        ax.set_xlim(x_dim*t0,x_dim*tf)
        ax.set_ylim(0,img_shape[1])
        for t in range(t0,tf):
            padded_img = np.pad(kymo_arr[t][::-1],((0,0),(self.viewpadding,self.viewpadding)),'constant')
            ax.imshow(padded_img, extent=[t*x_dim, (t+1)*x_dim, 0, img_shape[1]])

            merged_centroids = (centroids[t][:-1,:]+centroids[t][1:,:])/2

            adjusted_centroids = centroids[t] + t*np.array([x_dim,0]) + np.array([self.viewpadding,0])
            adjusted_merged_centroids = merged_centroids + t*np.array([x_dim,0]) + np.array([self.viewpadding,0])

            if t > t0:

                N_detect_i = len(centroids[t-1])
                N_detect_f = len(centroids[t])

                for i in range(N_detect_i-1):
                    for j in range(N_detect_f):

                        centroid_k1 = centroids[t-1][i]
                        centroid_k2 = centroids[t-1][i+1]

                        centroid_i = centroids[t][j]

                        color_score = abs(Cki[t-1][i,j]/-100)

                        ax.plot([centroid_k1[0] + (t-1)*x_dim + self.viewpadding, centroid_i[0] + t*x_dim + self.viewpadding],\
                                [(centroid_k1[1]+centroid_k2[1])/2,centroid_i[1]],c=cmap_fn(color_score),linewidth=3)

            ax.scatter(adjusted_centroids[:,0],adjusted_centroids[:,1],c="r",s=100,zorder=10)
            ax.scatter(adjusted_merged_centroids[:,0],adjusted_merged_centroids[:,1],c="white",s=100,zorder=10)

        merged_centroids = (centroids[tf][:-1,:]+centroids[tf][1:,:])/2

        adjusted_centroids = centroids[tf] + tf*np.array([x_dim,0]) + np.array([self.viewpadding,0])
        adjusted_merged_centroids = merged_centroids + tf*np.array([x_dim,0]) + np.array([self.viewpadding,0])

        ax.scatter(adjusted_centroids[:,0],adjusted_centroids[:,1],c="r",s=100,zorder=10)
        ax.scatter(adjusted_merged_centroids[:,0],adjusted_merged_centroids[:,1],c="white",s=100,zorder=10)
        
    def plot_score_metrics(self,kymo_df,trenchid,t_range=(0,-1),u_size=0.25,sig_size=0.05,u_pos=0.25,sig_pos=0.1,w_merge=0.8,viewpadding=0):
        self.viewpadding = viewpadding
        
        trench = kymo_df.loc[trenchid]
        file_idx = trench["File Index"].unique().tolist()[0]
        trench_idx = trench["File Trench Index"].unique().tolist()[0]
        orientation_dict = {"top":0,"bottom":1}
        orientation = trench["lane orientation"].unique().tolist()[0]
        orientation = orientation_dict[orientation]
                
        with h5py.File(self.segpath + "/segmentation_" + str(file_idx) + ".hdf5") as infile:
            data = infile["data"][trench_idx]
        
        t0,tf = t_range
        self.u_pos,self.sig_pos = (u_pos,sig_pos)
        self.u_size,self.sig_size = (u_size,sig_size)
        self.w_merge = w_merge
        labeled_data = get_labeled_data(data,orientation)
        
        centroids,sizes,bboxes,empty_trenches = get_segment_props(labeled_data,interpolate_empty_trenches=True)
        Cij,Cik,Cki,_,_,_,_ = self.compute_score_arrays(centroids,sizes,bboxes)
            
        fig1, (ax1, ax2) = plt.subplots(1, 2)
        fig1.set_size_inches(18, 4)
        
        
        self.plot_neighbor_stats(ax1,ax2,centroids,bboxes,sizes)
                
        fig2, axs = plt.subplots(4)
        fig2.set_size_inches(18, 24)
        
        self.plot_segmentation(axs[0],labeled_data,t0=t0,tf=tf,cmap='coolwarm')
        axs[0].set_title('Segmentation')
        self.plot_ij_scores(axs[1],labeled_data,centroids,Cij,t0=t0,tf=tf,cmap='coolwarm')
        axs[1].set_title('Movement Transition Score')
        self.plot_ik_scores(axs[2],labeled_data,centroids,Cik,t0=t0,tf=tf,cmap='coolwarm')
        axs[2].set_title('Division Transition Score')
        self.plot_ki_scores(axs[3],labeled_data,centroids,Cki,t0=t0,tf=tf,cmap='coolwarm')
        axs[3].set_title('Merge Transition Score')
        
        plt.show()
        
        return data,orientation,empty_trenches
        
    def interactive_scorefn(self):
        meta_handle = pandas_hdf5_handler(self.headpath+"/metadata.hdf5")
        kymo_df = meta_handle.read_df("kymograph")
        num_trenches = len(kymo_df.index.get_level_values(0).unique())
        timepoints = len(kymo_df.index.get_level_values(1).unique())
  
        self.output = interactive(self.plot_score_metrics,{"manual":True},\
                 kymo_df = fixed(kymo_df),trenchid=IntSlider(value=0, min=0, max=num_trenches-1, step=1),\
                 t_range=IntRangeSlider(value=[0, timepoints-1],min=0,max=timepoints-1,\
                 step=1,disabled=False,continuous_update=False),\
                 u_size=FloatSlider(value=self.u_size, min=0., max=1., step=0.01),\
                 sig_size=FloatSlider(value=self.sig_size, min=0., max=0.5, step=0.01),\
                 u_pos=FloatSlider(value=self.u_pos, min=0., max=1., step=0.01),\
                 sig_pos=FloatSlider(value=self.sig_pos, min=0., max=0.5, step=0.01),\
                 w_merge=FloatSlider(value=self.w_merge, min=0., max=1., step=0.01),\
                 viewpadding=IntSlider(value=0, min=0, max=20, step=1));
        
        display(self.output)

class tracking_solver:
    def __init__(self,headpath,segfolder,paramfile=False,ScoreFn=False,intensity_channel_list=None,props_list=['centroid','area'],\
                 props_to_unpack={'centroid':['centroid y','centroid x']},pixel_scaling_factors={'centroid y': 1,'centroid x': 1, 'area': 2},\
                 intensity_props_list=['mean_intensity'],merge_per_iter=0,conv_tolerence=2,edge_limit=3,\
                 u_size=0.22,sig_size=0.08,u_pos=0.21,sig_pos=0.1,w_merge=0.8):
        
                
        if paramfile:
            parampath = headpath + "/lineage_tracing.par"
            with open(parampath, 'rb') as infile:
                param_dict = pkl.load(infile)
            
            intensity_channel_list = param_dict["Channels:"]
            merge_per_iter = param_dict["Merges Detected Per Iteration:"]
            conv_tolerence = param_dict["# Unimproved Iterations Before Stopping:"]
            edge_limit = param_dict["Closest N Objects to Consider:"]
            u_size = param_dict["Mean Size Increase:"]
            sig_size = param_dict["Standard Deviation of Size Increase:"]
            u_pos = param_dict["Mean Position Increase:"]
            sig_pos = param_dict["Standard Deviation of Position Increase:"]
            w_merge = param_dict["Cell Merging Weight:"]
        
        self.headpath = headpath
        self.segpath = headpath + "/" + segfolder
        self.kymopath = headpath + "/kymograph"
        self.lineagepath = headpath + "/lineage"
        self.metapath = self.headpath + "/metadata.hdf5"
        self.meta_handle = pandas_hdf5_handler(self.metapath)
        fovdf = self.meta_handle.read_df("global",read_metadata=True)
        self.metadata = fovdf.metadata
        self.intensity_channel_list = intensity_channel_list
        self.props_list = props_list
        
        
        self.pixel_scaling_factors = pixel_scaling_factors
        self.props_to_unpack = props_to_unpack
        self.intensity_props_list = intensity_props_list
        
        if ScoreFn:
            self.ScoreFn = ScoreFn
        else:
            self.ScoreFn = scorefn(headpath,segfolder,u_size=u_size,sig_size=sig_size,u_pos=u_pos,sig_pos=sig_pos,w_merge=w_merge)
        self.merge_per_iter = merge_per_iter
        self.conv_tolerence = conv_tolerence
        self.edge_limit = edge_limit
        
        self.viewpadding = 0

    def define_tracking_problem(self,Cij,Cik,Cki,posij,posik,poski,sizes,merge_per_iter,edge_limit):

        prob = LpProblem("Lineage",LpMinimize)
        
        Aij = {}
        Aik = {}
        Aki = {}
        Ai = {}

        obj_fn_terms = []
        obj_fn_coeff = []
        
        valid_j_list = []
        
        for t in range(len(Cij)):
            cij = Cij[t]
            
            Aij[t] = {"ij":{}, "ji":{}}
            Aik[t] = {"ik":{}, "ki":{}}
            Aki[t] = {"ki":{}, "ik":{}}
            
            for i in range(cij.shape[0]):
                Aij[t]["ij"][i] = {}
                Aik[t]["ik"][i] = {}
                if i < (cij.shape[0]-1):
                    Aki[t]["ki"][i] = {}
                
            for j in range(cij.shape[1]):
                Aij[t]["ji"][j] = {}
                Aki[t]["ik"][j] = {}
                if j < (cij.shape[1]-1):
                    Aik[t]["ki"][j] = {}
                        
            working_posij = posij[t]    
            working_posik = posik[t]    
            working_poski = poski[t]
            
            valid_posij = np.argsort(-working_posij,axis=1)[:,:edge_limit]
            valid_posik = np.argsort(-working_posik,axis=1)[:,:edge_limit]
            valid_poski = np.argsort(-working_poski,axis=1)[:,:edge_limit]

            for i in range(cij.shape[0]):
                                
                valid_posij_slice = valid_posij[i]
                for j in range(cij.shape[1]):
                    if j >= i:
                        if j in valid_posij_slice:
                            var = LpVariable("ai="+str(i)+",j="+str(j)+",t="+str(t),0,1,cat='Continuous')
                            Aij[t]["ij"][i][j] = var
                            Aij[t]["ji"][j][i] = var                            
                            cijt = cij[i,j]
                            obj_fn_terms.append(var)
                            obj_fn_coeff.append(cijt)

                        if j < (cij.shape[1]-1):
                            valid_posik_slice = valid_posik[i]
                            if j in valid_posik_slice:
                                var = LpVariable("ai="+str(i)+",k="+str(j)+",t="+str(t),0,1,cat='Continuous')
                                Aik[t]["ik"][i][j] = var
                                Aik[t]["ki"][j][i] = var
                                cikt = Cik[t][i,j]
                                obj_fn_terms.append(var)
                                obj_fn_coeff.append(cikt)

                        if i < (cij.shape[0]-1):
                            valid_poski_slice = valid_poski[i]
                            if j in valid_poski_slice:                            
                                var = LpVariable("ak="+str(i)+",i="+str(j)+",t="+str(t),0,1,cat='Continuous')
                                Aki[t]["ki"][i][j] = var
                                Aki[t]["ik"][j][i] = var
                                ckit = Cki[t][i,j]
                                obj_fn_terms.append(var)
                                obj_fn_coeff.append(ckit)

                Ai[(i,t)] = LpVariable("ai="+str(i)+",t="+str(t),0,1,cat='Continuous')     
                obj_fn_terms.append(Ai[(i,t)])
                obj_fn_coeff.append(0.)

        prob += lpSum([obj_fn_coeff[n]*obj_fn_term for n,obj_fn_term in enumerate(obj_fn_terms)]), "Objective Function"
        
        for t in range(0,len(sizes)):
            if t == 0:
                N_detect = len(sizes[t])
                
                for j in range(N_detect):
                    neg_Aij = list(Aij[t]["ij"].get(j,{}).values())
                    neg_Aik = list(Aik[t]["ik"].get(j,{}).values())
                    neg_Aki = list(Aki[t]["ki"].get(j-1,{}).values()) + list(Aki[t]["ki"].get(j,{}).values())
                    neg_Ai = Ai[(j,t)]

                    prob += lpSum(neg_Aij) + lpSum(neg_Aik) + lpSum(neg_Aki) + neg_Ai <= 1,\
                    "Constraint 7 (" + str(t) + "," + str(j) + ")"

            elif t < len(sizes)-1:
                N_detect_i = len(sizes[t-1])
                N_detect_f = len(sizes[t])

                for j in range(N_detect_f):
                    pos_Aij = list(Aij[t-1]["ji"].get(j,{}).values())                    
                    pos_Aik = list(Aik[t-1]["ki"].get(j-1,{}).values()) + list(Aik[t-1]["ki"].get(j,{}).values())
                    pos_Aki = list(Aki[t-1]["ik"].get(j,{}).values())
                    
                    
                    neg_Aij = list(Aij[t]["ij"].get(j,{}).values())
                    neg_Aik = list(Aik[t]["ik"].get(j,{}).values())
                    neg_Aki = list(Aki[t]["ki"].get(j-1,{}).values()) + list(Aki[t]["ki"].get(j,{}).values())
                    neg_Ai = Ai[(j,t)]                
                    
                    prob += lpSum(pos_Aij) + lpSum(pos_Aik) + lpSum(pos_Aki) - \
                        lpSum(neg_Aij) - lpSum(neg_Aik) - lpSum(neg_Aki) - neg_Ai == 0,\
                        "Constraint 9 (" + str(t) + "," + str(j) + ")"

                    prob += lpSum(pos_Aij) + lpSum(pos_Aik) + lpSum(pos_Aki) <= 1,\
                    "Constraint 6 (" + str(t) + "," + str(j) + ")"

                    prob += lpSum(neg_Aij) + lpSum(neg_Aik) + lpSum(neg_Aki) + neg_Ai <= 1,\
                        "Constraint 7 (" + str(t) + "," + str(j) + ")"

                    if j != N_detect_f - 1:

                        i_above = list(range(j+1,N_detect_f))

                        above_Aij = []
                        above_Aik = []
                        above_Aki = list(Aki[t]["ki"].get(j,{}).values())
                        above_Ai = []
                        for i in i_above:                            
                            above_Aij += list(Aij[t]["ij"].get(i,{}).values())
                            above_Aik += list(Aik[t]["ik"].get(i,{}).values())
                            above_Aki += list(Aki[t]["ki"].get(i,{}).values())
                            above_Ai.append((i,t))

                        prob += N_detect_f*neg_Ai + lpSum(above_Aij) + \
                        lpSum(above_Aik) + lpSum(above_Aki) <= N_detect_f,\
                        "Contraint 8 (" + str(t) + "," + str(j) + ")"

            else:
                N_detect = len(sizes[t])

                for j in range(N_detect):
                    pos_Aij = list(Aij[t-1]["ji"].get(j,{}).values())                    
                    pos_Aki = list(Aki[t-1]["ik"].get(j,{}).values())
                    pos_Aik = list(Aik[t-1]["ki"].get(j-1,{}).values()) + list(Aik[t-1]["ki"].get(j,{}).values())

                    if j == N_detect_f - 1:

                        prob += lpSum(pos_Aij) + lpSum(pos_Aik) + \
                        lpSum(pos_Aki) <= 1, "Constraint 6 (" + str(t) + "," + str(j) + ")"

                    else:

                        prob += lpSum(pos_Aij) + lpSum(pos_Aik) + \
                        lpSum(pos_Aki) <= 1, "Constraint 6 (" + str(t) + "," + str(j) + ")"
        
        all_Aki_vars = []
        for _,tpt in Aki.items():
            for _,subdict in tpt["ki"].items():
                all_Aki_vars += list(subdict.values())
                                        
        prob += lpSum(all_Aki_vars) <= merge_per_iter, "Merge Limit"

        return prob,Aij,Aik,Aki,Ai

    def get_sln_params(self,Aij,Aik,Aki,Ai,posij,posik,poski,sizes,edge_limit):
        Aij_arr_list = [np.zeros((len(sizes[t]),len(sizes[t+1])),dtype=bool) for t in range(len(sizes)-1)]
        Aik_arr_list = [np.zeros((len(sizes[t]),len(sizes[t+1])-1),dtype=bool) for t in range(len(sizes)-1)]
        Aki_arr_list = [np.zeros((len(sizes[t])-1,len(sizes[t+1])),dtype=bool) for t in range(len(sizes)-1)]
        Ai_arr_list = [np.zeros(len(sizes[t]),dtype=bool) for t in range(len(sizes))]

        for t in range(len(sizes)-1):
            
            working_posij = posij[t]    
            working_posik = posik[t]    
            working_poski = poski[t]
            
            valid_posij = np.argsort(-working_posij,axis=1)[:,:edge_limit]
            valid_posik = np.argsort(-working_posik,axis=1)[:,:edge_limit]
            valid_poski = np.argsort(-working_poski,axis=1)[:,:edge_limit]
            
            num_i = Aij_arr_list[t].shape[0]
            num_j = Aij_arr_list[t].shape[1]
            
            for i in range(num_i):
                valid_posij_slice = valid_posij[i]
                for j in range(num_j):
                    if j >= i:
                        if j in valid_posij_slice:                            
                            Aij_arr_list[t][i,j] = bool(np.round(Aij[t]["ij"][i][j].varValue))

                        if j < num_j - 1:
                            valid_posik_slice = valid_posik[i]
                            if j in valid_posik_slice:
                                Aik_arr_list[t][i,j] = bool(np.round(Aik[t]["ik"][i][j].varValue))

                        if i < num_i - 1:
                            valid_poski_slice = valid_poski[i]
                            if j in valid_poski_slice:
                                Aki_arr_list[t][i,j] = bool(np.round(Aki[t]["ki"][i][j].varValue))

                Ai_arr_list[t][i] = bool(np.round(Ai[(i,t)].varValue))

        return Aij_arr_list,Aik_arr_list,Aki_arr_list,Ai_arr_list

    def split_all_merged(self,labeled_data,Aki_arr_list):
        new_labeled_data = copy.copy(labeled_data)
        for t in range(len(Aki_arr_list)):
            if np.any(Aki_arr_list[t]):
                merged_label_list = (np.where(np.any(Aki_arr_list[t],axis=1))[0]+1).tolist()
                output_arr = self.split_cells(labeled_data[t+1],merged_label_list)
                new_labeled_data[t+1] = output_arr
        return new_labeled_data

    def split_cells(self,labeled_arr,merged_label_list):
        output_arr = copy.copy(labeled_arr)
        current_idx = 1
        for idx in range(1,np.max(labeled_arr)+1):

            if idx in merged_label_list:

                bool_arr = (output_arr==idx)
                cell_indices = np.where(bool_arr)
                X = np.array(list(zip(cell_indices[0],cell_indices[1])))
                kmeans = skl.cluster.KMeans(n_clusters=2, random_state=0)

                C = kmeans.fit_predict(X)
                centroids = kmeans.cluster_centers_.astype(int)
                C0 = (C==0)
                zero_indices = (cell_indices[0][C0],cell_indices[1][C0])
                one_indices = (cell_indices[0][~C0],cell_indices[1][~C0])

                zero_first = centroids[0,0] < centroids[1,0]
                if zero_first:
                    output_arr[zero_indices] = current_idx
                    output_arr[one_indices] = current_idx+1
                else:
                    output_arr[zero_indices] = current_idx+1
                    output_arr[one_indices] = current_idx

                current_idx += 2

            else:
                output_arr[labeled_arr==idx] = current_idx

                current_idx += 1

        return output_arr
    
    def plot_tracking_sln(self,kymo_arr,centroids,Aij_arr_list,Aik_arr_list,Aki_arr_list,t0=0,tf=-1,x_size=18,y_size=6):
        
        img_shape = kymo_arr.shape[1:][::-1]
        x_dim = img_shape[0]+(self.viewpadding*2)
        if tf == -1:
            tf = len(centroids)-1
        
        fig, ax = plt.subplots(1)
        fig.set_size_inches(x_size, y_size)
        
        ax.set_xlim(x_dim*t0,x_dim*tf)
        ax.set_ylim(0,img_shape[1])
        for t in range(t0,tf):
            padded_img = np.pad(kymo_arr[t][::-1],((0,0),(self.viewpadding,self.viewpadding)),'constant')
            ax.imshow(np.ma.array(padded_img, mask=(padded_img == 0)), extent=[t*x_dim, (t+1)*x_dim, 0, img_shape[1]], cmap="Set1", vmin=0)
            
            adjusted_centroids = centroids[t] + t*np.array([x_dim,0]) + np.array([self.viewpadding,0])
            ax.scatter(adjusted_centroids[:,0],adjusted_centroids[:,1],c="r")

            Aij_cords = np.where(Aij_arr_list[t])
            for idx in range(len(Aij_cords[0])):
                i = Aij_cords[0][idx]
                j = Aij_cords[1][idx]
                centroid_i = centroids[t][i]
                centroid_j = centroids[t+1][j]

                ax.plot([centroid_i[0] + t*x_dim + self.viewpadding, centroid_j[0] + (t+1)*x_dim + self.viewpadding],[centroid_i[1],centroid_j[1]],c="c",linewidth=3)

            Aik_cords = np.where(Aik_arr_list[t])
            for idx in range(len(Aik_cords[0])):
                i = Aik_cords[0][idx]
                k = Aik_cords[1][idx]
                centroid_i = centroids[t][i]

                centroid_k1 = centroids[t+1][k]
                centroid_k2 = centroids[t+1][k+1]

                ax.plot([centroid_i[0] + t*x_dim + self.viewpadding, centroid_k1[0] + (t+1)*x_dim + self.viewpadding],[centroid_i[1],centroid_k1[1]],c="c",linewidth=3)
                ax.plot([centroid_i[0] + t*x_dim + self.viewpadding, centroid_k2[0] + (t+1)*x_dim + self.viewpadding],[centroid_i[1],centroid_k2[1]],c="c",linewidth=3)

            Aki_cords = np.where(Aki_arr_list[t])
            for idx in range(len(Aki_cords[0])):
                k = Aki_cords[0][idx]
                i = Aki_cords[1][idx]
                centroid_k1 = centroids[t][k]
                centroid_k2 = centroids[t][k+1]

                centroid_i = centroids[t+1][i]

                ax.plot([centroid_k1[0] + t*x_dim + self.viewpadding, centroid_i[0] + (t+1)*x_dim + self.viewpadding],[centroid_k1[1],centroid_i[1]],c="c",linewidth=3)
                ax.plot([centroid_k2[0] + t*x_dim + self.viewpadding, centroid_i[0] + (t+1)*x_dim + self.viewpadding],[centroid_k2[1],centroid_i[1]],c="c",linewidth=3)

        padded_img = np.pad(kymo_arr[tf][::-1],((0,0),(self.viewpadding,self.viewpadding)),'constant')
        ax.imshow(padded_img, extent=[tf*x_dim, (tf+1)*x_dim, 0, img_shape[1]])
        adjusted_centroids = centroids[tf] + tf*np.array([x_dim,0])    
        ax.scatter(adjusted_centroids[:,0],adjusted_centroids[:,1],c="r")
        
        plt.show()
        
    def plot_cleaned_sln(self,kymo_arr,centroids,Aij_arr_list,Aik_arr_list,Aki_arr_list,t0=0,tf=-1,x_size=18,y_size=6):
        
        img_shape = kymo_arr.shape[1:][::-1]
        x_dim = img_shape[0]+(self.viewpadding*2)
        if tf == -1:
            tf = len(centroids)-1
        
        fig, ax = plt.subplots(1)
        fig.set_size_inches(x_size, y_size)
        
        ax.set_xlim(x_dim*t0,x_dim*tf)
        ax.set_ylim(0,img_shape[1])
        for t in range(t0,tf):
            padded_img = np.pad(kymo_arr[t][::-1],((0,0),(self.viewpadding,self.viewpadding)),'constant')
            ax.imshow(np.ma.array(padded_img, mask=(padded_img == 0)), extent=[t*x_dim, (t+1)*x_dim, 0, img_shape[1]], cmap="Set1", vmin=0)
            
            adjusted_centroids = centroids[t] + t*np.array([x_dim,0]) + np.array([self.viewpadding,0])
            ax.scatter(adjusted_centroids[:,0],adjusted_centroids[:,1],c="r")
            
            Aik_cords = np.where(Aik_arr_list[t])
            div_is = Aik_cords[0]
            working_j = 0
            
            for i in range(len(centroids[t])):
                centroid_i = centroids[t][i]
                
                if working_j > len(centroids[t+1])-1:
                    break
                
                if i in div_is:
                    
                    if working_j > len(centroids[t+1])-2:
                        centroid_k1 = centroids[t+1][working_j]
                        ax.plot([centroid_i[0] + t*x_dim + self.viewpadding, centroid_k1[0] + (t+1)*x_dim + self.viewpadding],[centroid_i[1],centroid_k1[1]],c="c",linewidth=3)
                        working_j += 1
                        
                    else:
                        centroid_k1 = centroids[t+1][working_j]
                        centroid_k2 = centroids[t+1][working_j+1]
                        ax.plot([centroid_i[0] + t*x_dim + self.viewpadding, centroid_k1[0] + (t+1)*x_dim + self.viewpadding],[centroid_i[1],centroid_k1[1]],c="c",linewidth=3)
                        ax.plot([centroid_i[0] + t*x_dim + self.viewpadding, centroid_k2[0] + (t+1)*x_dim + self.viewpadding],[centroid_i[1],centroid_k2[1]],c="c",linewidth=3)
                        working_j += 2
                else:
                    centroid_j = centroids[t+1][working_j]
                    ax.plot([centroid_i[0] + t*x_dim + self.viewpadding, centroid_j[0] + (t+1)*x_dim + self.viewpadding],[centroid_i[1],centroid_j[1]],c="c",linewidth=3)
                    working_j += 1

        padded_img = np.pad(kymo_arr[tf][::-1],((0,0),(self.viewpadding,self.viewpadding)),'constant')
        ax.imshow(padded_img, extent=[tf*x_dim, (tf+1)*x_dim, 0, img_shape[1]])
        adjusted_centroids = centroids[tf] + tf*np.array([x_dim,0]) + np.array([self.viewpadding,0])
        ax.scatter(adjusted_centroids[:,0],adjusted_centroids[:,1],c="r")
        
        plt.show()
        
    def run_iteration(self,labeled_data,merge_per_iter,edge_limit):
        centroids,sizes,bboxes,empty_trenches = get_segment_props(labeled_data,interpolate_empty_trenches=True)
        Cij,Cik,Cki,_,posij,posik,poski = self.ScoreFn.compute_score_arrays(centroids,sizes,bboxes)
        prob,Aij,Aik,Aki,Ai = self.define_tracking_problem(Cij,Cik,Cki,posij,posik,poski,sizes,merge_per_iter,edge_limit)  
        prob.solve()
        Aij_arr_list,Aik_arr_list,Aki_arr_list,Ai_arr_list = self.get_sln_params(Aij,Aik,Aki,Ai,posij,posik,poski,sizes,edge_limit)

        return centroids,sizes,prob,Aij_arr_list,Aik_arr_list,Aki_arr_list,Ai_arr_list,empty_trenches
    
    def compute_lineage(self,data,orientation):
        labeled_data = get_labeled_data(data,orientation)
        working_labeled_data = copy.copy(labeled_data)

        iter_outputs = []
        iter_scores = []
        iter_labeled_data = []
        iter_labeled_data.append(labeled_data)

        iter_outputs.append(self.run_iteration(working_labeled_data,self.merge_per_iter,self.edge_limit))

        centroids,sizes,prob,Aij_arr_list,Aik_arr_list,Aki_arr_list,Ai_arr_list,empty_trenches = iter_outputs[-1]
        
        merged_cells = np.any([np.any(Aki_arr) for Aki_arr in Aki_arr_list])
        objective_val = value(prob.objective)
        active_edges = sum([v.varValue for v in prob.variables()])
        edge_normalized_objective = objective_val/active_edges

        print("Objective = ", objective_val)
        print("Number of Active Edges = ", active_edges)
        print("Edge Normalized Objective = ", edge_normalized_objective)
        
        if self.merge_per_iter == 0:
            
            return working_labeled_data,centroids,sizes,Aij_arr_list,Aik_arr_list,Aki_arr_list,Ai_arr_list,edge_normalized_objective,empty_trenches

        iter_scores.append(edge_normalized_objective)
        iter_since_incr = 0

        while merged_cells:
            working_labeled_data = self.split_all_merged(working_labeled_data,Aki_arr_list)
            iter_labeled_data.append(working_labeled_data)
            iter_outputs.append(self.run_iteration(working_labeled_data,self.merge_per_iter,self.edge_limit))
            _,_,prob,_,_,Aki_arr_list,_,_ = iter_outputs[-1]

            merged_cells = np.any([np.any(Aki_arr) for Aki_arr in Aki_arr_list])

            objective_val = value(prob.objective)
            active_edges = sum([v.varValue for v in prob.variables()])
            edge_normalized_objective = objective_val/active_edges

            print("Objective = ", objective_val)
            print("Number of Active Edges = ", active_edges)
            print("Edge Normalized Objective = ", edge_normalized_objective)

            iter_scores.append(edge_normalized_objective)
            min_iter_score = min(iter_scores)
            
            if iter_scores[-1]<=min_iter_score:
                iter_since_incr = 0
            else:
                iter_since_incr += 1
            
            if iter_since_incr>self.conv_tolerence:
                break

        best_iter_idx = np.argmin(iter_scores)
        working_labeled_data = iter_labeled_data[best_iter_idx]
        centroids,sizes,prob,Aij_arr_list,Aik_arr_list,Aki_arr_list,Ai_arr_list,empty_trenches = self.run_iteration(working_labeled_data,0,self.edge_limit)
        objective_val = value(prob.objective)
        active_edges = sum([v.varValue for v in prob.variables()])
        edge_normalized_objective = objective_val/active_edges

        return working_labeled_data,centroids,sizes,Aij_arr_list,Aik_arr_list,Aki_arr_list,Ai_arr_list,edge_normalized_objective,empty_trenches
    
    def get_nuclear_lineage(self,sizes,empty_trenches,Aik_arr_list):
        
        cell_ids = list(range(len(sizes[0])))
        current_max_id = np.max(cell_ids)
        cell_ids_list = [copy.copy(cell_ids)]
        mother_dict = {}
        daughter_dict = {}
        sister_dict = {}

        for t in range(0,len(Aik_arr_list)):
            max_cells = len(sizes[t+1])
            if t in empty_trenches:
                cell_ids_list.append(copy.copy(cell_ids_list[-1]))
            else:
                Aik_cords = np.where(Aik_arr_list[t])
                ttl_added = 0
                for idx in range(len(Aik_cords[0])):
                    i = Aik_cords[0][idx]
                    current_idx = i+ttl_added
                    
                    if current_idx == (max_cells-1):
                        
                        daughter_id1,daughter_id2 = (current_max_id+1,-1)

                        mother_dict[daughter_id1] = cell_ids[current_idx]
                        sister_dict[daughter_id1] = daughter_id2
                        
                        daughter_dict[cell_ids[current_idx]] = (daughter_id1,daughter_id2)

                        del cell_ids[current_idx]
                        cell_ids.insert(current_idx,daughter_id1)

                        current_max_id += 1
                        break
                        
                    elif current_idx>(max_cells-1):
                        daughter_id1,daughter_id2 = (-1,-1)
                        daughter_dict[cell_ids[current_idx]] = (daughter_id1,daughter_id2)
                        del cell_ids[current_idx]
                        break
                        
                    else:
                    
                        daughter_id1,daughter_id2 = (current_max_id+1,current_max_id+2)

                        mother_dict[daughter_id1] = cell_ids[current_idx]
                        sister_dict[daughter_id1] = daughter_id2
                        
                        mother_dict[daughter_id2] = cell_ids[current_idx]
                        sister_dict[daughter_id2] = daughter_id1
                        
                        daughter_dict[cell_ids[current_idx]] = (daughter_id1,daughter_id2)

                        del cell_ids[current_idx]
                        cell_ids.insert(current_idx,daughter_id1)
                        cell_ids.insert(current_idx+1,daughter_id2)

                        current_max_id += 2
                        ttl_added += 1
                    
                cell_ids = copy.copy(cell_ids[:max_cells])
                cell_ids_list.append(copy.copy(cell_ids[:max_cells]))
                
        return mother_dict,daughter_dict,sister_dict,cell_ids_list
    
    def get_cell_lineage(self,file_idx,file_trench_idx,cell_id,mother_dict,daughter_dict,sister_dict):
        
        if cell_id in daughter_dict.keys():
            daughter_id_1 = daughter_dict[cell_id][0]
            daughter_id_2 = daughter_dict[cell_id][1]
        else:
            daughter_id_1 = -1
            daughter_id_2 = -1
        if cell_id in mother_dict.keys():
            mother_id = mother_dict[cell_id]
        else:
            mother_id = -1
        if cell_id in sister_dict.keys():
            sister_id = sister_dict[cell_id]
        else:
            sister_id = -1
        
        output_list = [mother_id,daughter_id_1,daughter_id_2,sister_id]
        output = []
        
        for item in output_list:
            if item == -1:
                output.append(-1)
            else:
                global_id = int(f'{file_idx:04}{file_trench_idx:04}{item:04}')
                output.append(global_id)

        return output
    
    def get_lineage_df(self,file_idx,file_trench_idx,lineage_score,labeled_data,empty_trenches,mother_dict,daughter_dict,sister_dict,cell_ids_list):

        kymograph_file = self.kymopath + "/kymograph_" + str(file_idx) + ".hdf5"
        if self.intensity_channel_list is not None:
            kymo_arr_list = []
            with h5py.File(kymograph_file,"r") as kymofile:
                for intensity_channel in self.intensity_channel_list:
                    kymo_arr_list.append(kymofile[intensity_channel][:])
                    
        pixel_microns = self.metadata['pixel_microns']
                
        props_output = []
        for t in range(labeled_data.shape[0]):
            if t not in empty_trenches:
                cell_ids = cell_ids_list[t]
                ttl_ids = len(cell_ids)
                
                #non intensity info first
                non_intensity_rps = sk.measure.regionprops(labeled_data[t])
                if self.intensity_channel_list is not None:
                    intensity_rps_list = []
                    for i,intensity_channel in enumerate(self.intensity_channel_list):
                        intensity_rps = sk.measure.regionprops(labeled_data[t], kymo_arr_list[i][file_trench_idx,t])
                        intensity_rps_list.append(intensity_rps)
                
                for idx in range(ttl_ids):
                    cell_id = cell_ids[idx]                    
                    global_cell_id = int(f'{file_idx:04}{file_trench_idx:04}{cell_id:04}')
                    rp = non_intensity_rps[idx]

                    props_entry = [file_idx, file_trench_idx, t, cell_id, global_cell_id, lineage_score]
                    props_entry += self.get_cell_lineage(file_idx,file_trench_idx,cell_id,mother_dict,daughter_dict,sister_dict)
                    
                    for prop_key in self.props_list:
                        prop = getattr(rp, prop_key)
                        if prop_key in self.props_to_unpack.keys():
                            prop_out = dict(zip(self.props_to_unpack[prop_key],list(prop)))
                        else:
                            prop_out = {prop_key:prop}
                        for key,value in prop_out.items():
                            if key in self.pixel_scaling_factors.keys():
                                output = value*(pixel_microns**self.pixel_scaling_factors[key])
                            else:
                                output = value
                            props_entry.append(output)
                                            
                    if self.intensity_channel_list is not None:
                        for i,intensity_channel in enumerate(self.intensity_channel_list):
                            intensity_rps=intensity_rps_list[i]
                            inten_rp = intensity_rps[idx]
                            
                            for prop_key in self.intensity_props_list:
                                prop = getattr(inten_rp, prop_key)
                                if prop_key in self.props_to_unpack.keys():
                                    prop_out = dict(zip(self.props_to_unpack[prop_key],list(prop)))
                                else:
                                    prop_out = {prop_key:prop}
                                for key,value in prop_out.items():
                                    if key in self.pixel_scaling_factors.keys():
                                        output = value*(pixel_microns**self.pixel_scaling_factors[key])
                                    else:
                                        output = value
                                    props_entry.append(output)
                                    
                    props_output.append(props_entry)
        
        base_list = ['File Index','File Trench Index','timepoints','CellID','Global CellID','Trench Score','Mother CellID','Daughter CellID 1','Daughter CellID 2','Sister CellID']

        unpacked_props_list = []
        for prop_key in self.props_list:
            if prop_key in self.props_to_unpack.keys():
                unpacked_names = self.props_to_unpack[prop_key]
                unpacked_props_list += unpacked_names
            else:
                unpacked_props_list.append(prop_key)
                
        if self.intensity_channel_list is not None:
            for channel in self.intensity_channel_list:
                for prop_key in self.intensity_props_list:
                    if prop_key in self.props_to_unpack.keys():
                        unpacked_names = self.props_to_unpack[prop_key]
                        unpacked_props_list += [channel + " " + item for item in unpacked_names]
                    else:
                        unpacked_props_list.append(channel + " " + prop_key)
                        
        column_list = base_list + unpacked_props_list
            
        df_out = pd.DataFrame(props_output, columns=column_list).reset_index()
        df_out = df_out.set_index(['File Index','File Trench Index','timepoints','CellID'], drop=True, append=False, inplace=False)

        df_out = df_out.sort_index()
        
        return df_out
    

    def lineage_trace(self,file_idx,file_trench_idx):
        meta_handle = pandas_hdf5_handler(self.headpath+"/metadata.hdf5")
        
        kymo_df = meta_handle.read_df("kymograph")
        fov_df = kymo_df.reset_index(inplace=False)
        fov_df = fov_df.set_index(["File Index","File Trench Index"], drop=True, append=False, inplace=False)
        fov_df = fov_df.sort_index()
        trench = fov_df.loc[file_idx,file_trench_idx]
        orientation_dict = {"top":0,"bottom":1}
        orientation = trench["lane orientation"].unique().tolist()[0]
        orientation = orientation_dict[orientation]
        
        with h5py.File(self.segpath + "/segmentation_" + str(file_idx) + ".hdf5") as infile:
            data = infile["data"][file_trench_idx]
        
        labeled_data,_,sizes,_,Aik_arr_list,_,_,lineage_score,empty_trenches = self.compute_lineage(data,orientation)
        
        mother_dict,daughter_dict,sister_dict,cell_ids_list = self.get_nuclear_lineage(sizes,empty_trenches,Aik_arr_list)
                
        df_out = self.get_lineage_df(file_idx,file_trench_idx,lineage_score,labeled_data,empty_trenches,mother_dict,daughter_dict,sister_dict,cell_ids_list)
        
        return df_out

    def lineage_trace_file(self,file_idx):
        writedir(self.lineagepath,overwrite=False)
        
        kymo_meta = self.meta_handle.read_df("kymograph")
        filedf = kymo_meta.reset_index(inplace=False)
        filedf = filedf.set_index(["File Index","File Trench Index"], drop=True, append=False, inplace=False)
        filedf = filedf.sort_index()
        filedf = filedf.loc[file_idx]
        trench_idx_list = filedf.index.get_level_values(0).unique().tolist()
        
        mergeddf = []
        for file_trench_idx in trench_idx_list:
            try:
                df_out = self.lineage_trace(file_idx,file_trench_idx)
                mergeddf.append(df_out)
            except:
                pass
        mergeddf = pd.concat(mergeddf)
        
        kymo_meta = self.meta_handle.read_df("kymograph")
        kymo_meta = kymo_meta.reset_index(inplace=False)
        kymo_meta = kymo_meta.set_index(["File Index","File Trench Index","timepoints"], drop=True, append=False, inplace=False)
        kymo_meta = kymo_meta.sort_index()
        kymo_meta = kymo_meta.loc[file_idx:file_idx]
        
        mergeddf = mergeddf.reset_index(inplace=False)
        mergeddf = mergeddf.set_index(["File Index","File Trench Index","timepoints"], drop=True, append=False, inplace=False)
        mergeddf = mergeddf.sort_index()        
        mergeddf = mergeddf.join(kymo_meta)
        mergeddf = mergeddf.reset_index(inplace=False)
        mergeddf = mergeddf.set_index(["File Index","File Trench Index","timepoints","CellID"], drop=True, append=False, inplace=False)
        del mergeddf["index"]
        mergeddf = mergeddf.sort_index()
        
        new_index = mergeddf.index.map(lambda x: int(f'{x[0]:04}{x[1]:04}{x[2]:04}{x[3]:04}'))
        mergeddf = mergeddf.reset_index(drop=False,inplace=False)
        mergeddf.index = [item for item in new_index]
        mergeddf = mergeddf.dropna()
        
        df_path = self.lineagepath + "/block_" + str(file_idx) + ".parquet"
        
        del kymo_meta
        
        mergeddf.to_parquet(df_path,engine='fastparquet',compression='gzip')
        print("Done.")
        
        return file_idx
            
    def lineage_trace_all_files(self,dask_cont):
        kymo_meta = self.meta_handle.read_df("kymograph")
        file_list = kymo_meta["File Index"].unique().tolist()
        num_file_jobs = len(file_list)

        random_priorities = np.random.uniform(size=(num_file_jobs,))
        for k,file_idx in enumerate(file_list):
            priority = random_priorities[k]
            future = dask_cont.daskclient.submit(self.lineage_trace_file,file_idx,retries=1,priority=priority)
            dask_cont.futures["File Index: " + str(file_idx)] = future
            
    def wait_for_completion(self,dask_cont):
        kymo_meta = self.meta_handle.read_df("kymograph")
        file_list = kymo_meta["File Index"].unique().tolist()
        num_file_jobs = len(file_list)
        file_idx_list = dask_cont.daskclient.gather([dask_cont.futures["File Index: " + str(file_idx)] for file_idx in file_list],errors="skip")
        
    def compute_all_lineages(self,dask_cont,n_workers=100,memory='4GB',walltime='04:00:00',local=False):        
        writedir(self.lineagepath,overwrite=True)
        dask_cont.daskclient.restart()
        dask_cont.futures = {}
        try:
            self.lineage_trace_all_files(dask_cont)
        except:
            raise

    def test_tracking(self,data,orientation,intensity_channel_list=None,t_range=(0,-1),merge_per_iter=0,edge_limit=3,conv_tolerence=2,viewpadding=0,x_size=18,y_size=6):
        if len(intensity_channel_list) == 0:
            self.intensity_channel_list = None
        else:
            self.intensity_channel_list = list(intensity_channel_list)
        self.viewpadding = viewpadding
        self.merge_per_iter = merge_per_iter
        self.edge_limit = edge_limit
        self.conv_tolerence = conv_tolerence
        
        labeled_data,centroids,sizes,Aij_arr_list,Aik_arr_list,Aki_arr_list,Ai_arr_list,lineage_score,empty_trenches = self.compute_lineage(data,orientation)
    
        self.plot_cleaned_sln(labeled_data,centroids,Aij_arr_list,Aik_arr_list,Aki_arr_list,t0=t_range[0],tf=t_range[1],x_size=x_size,y_size=y_size)
    
    def interactive_tracking(self,data,orientation):
        
        available_channels = self.metadata["channels"]
        
        output = interactive(self.test_tracking,{"manual":True},data=fixed(data),orientation=fixed(orientation),\
                 intensity_channel_list = SelectMultiple(options=available_channels),\
                 t_range=IntRangeSlider(value=[0, data.shape[0]-1],min=0,max=data.shape[0]-1,\
                 step=1,disabled=False,continuous_update=False),\
                 merge_per_iter=IntSlider(value=self.merge_per_iter, min=0, max=10, step=1),\
                 edge_limit=IntSlider(value=self.edge_limit, min=1, max=6, step=1),\
                 conv_tolerence=IntSlider(value=self.conv_tolerence, min=1, max=4, step=1),\
                 viewpadding=IntSlider(value=0, min=0, max=20, step=1),\
                 x_size=IntSlider(value=18, min=0, max=30, step=1),\
                 y_size=IntSlider(value=8, min=0, max=20, step=1));
        
        display(output)
        
    def save_params(self):
        param_dict = {}
        
        param_dict["Channels:"] = self.intensity_channel_list
        param_dict["Merges Detected Per Iteration:"] = self.merge_per_iter
        param_dict["# Unimproved Iterations Before Stopping:"] = self.conv_tolerence
        param_dict["Closest N Objects to Consider:"] = self.edge_limit
        param_dict["Mean Size Increase:"] = self.ScoreFn.u_size
        param_dict["Standard Deviation of Size Increase:"] = self.ScoreFn.sig_size
        param_dict["Mean Position Increase:"] = self.ScoreFn.u_pos
        param_dict["Standard Deviation of Position Increase:"] = self.ScoreFn.sig_pos
        param_dict["Cell Merging Weight:"] = self.ScoreFn.w_merge
        
        for key,value in param_dict.items():
            print(key + " " + str(value))
                        
        with open(self.headpath + "/lineage_tracing.par", "wb") as outfile:
            pkl.dump(param_dict, outfile)


import numpy as np
from matplotlib import pyplot as plt

def plot_kymograph(kymograph):
    """Helper function for plotting kymographs. Takes a kymograph array of shape (y_dim,x_dim,t_dim).
    
    Args:
        kymograph (array): kymograph array of shape (y_dim,x_dim,t_dim).
    """
    list_in_t = [kymograph[t,:,:] for t in range(kymograph.shape[0])]
    img_arr = np.concatenate(list_in_t,axis=1)
    plt.imshow(img_arr)


import cv2
import os
import numpy as np
import matplotlib.pyplot as plt
from skimage.util import pad

def find_seed_image(image_stack, max_poi_std=5):
    """ Find a suitable image for segmentation from an image stack by checking interest point
    matching to neighbouring timepoints (out of focus or very drifted timepoints should not
    get good matches). We measure this by looking at the standard deviation of interest point
    matches.

    Args:
        image_stack (numpy.ndarray, t x y x x): Image stack
        max_poi_std: Standard deviation in Euclidean distance between interest points
    Returns:
        seed_index (int): Timepoint to use as seed
    
    """
    num_points = image_stack.shape[0]
    for i in range(num_points-1):
        # Get standard deviation of interest point matches
        _, _, std_distance = get_orb_pois(image_stack[i], image_stack[i+1])
        # If it's suitable use this as a seed
        if std_distance < max_poi_std:
            seed_index = i
            break
        if i == num_points-2:
            raise Exception("Could not find a suitable seed image in the timeseries")
    return seed_index

def scale_image(im):
    """ Stretch image intensities to fill whole histogram and convert to 8bit for opencv

    Args:
        im: image to scale
    Returns:
        im_rescaled: scaled image

    """

    # Get min and max
    min_intensity = np.min(im, axis=None)
    max_intensity = np.max(im, axis=None)
    # Stretch histogram
    scaling_factor = 255/(max_intensity - min_intensity)
    im_rescaled = (im-min_intensity)*scaling_factor
    # Convert to 8bit
    im_rescaled = im_rescaled.astype(np.uint8)
    return im_rescaled

def get_orb_pois(im1, im2, max_interest_points= 750, match_threshold=0.1, plot=False, pad_width=15):
    """ Find matching points of interest between two images according to ORB algorithm

    Args:
        im1 (numpy.ndarray, 8-bit integer): first image (y x x)
        im2 (numpy.ndarray, 8-bit integer): second image (y x x)
        max_interest_points (int): Maximum number of points to recognize using ORB algorithm
        match_threshold (float): The fraction of closest matches in Euclidean distance 
    Returns:
        points1 (numpy.ndarray, int): List of filtered points in image 1 that matched image 2 (n x 2)
        points2 (numpy.ndarray, int): List of filtered points in image 2 that matched image 1 (n x 2)
        std_distance (float): Standard deviation of distances between the matched points
    """
    orb = cv2.ORB_create(max_interest_points, patchSize=31)
    # Scale images
    im1_scaled = scale_image(im1)
    im2_scaled = scale_image(im2)
    # Pad images so that interest points on edges can be recognized
    im1_scaled = pad(im1_scaled, pad_width=pad_width, mode='reflect')
    im2_scaled = pad(im2_scaled, pad_width=pad_width, mode='reflect')
    # Match points based on their ORB descriptors
    matcher = cv2.DescriptorMatcher_create(cv2.DESCRIPTOR_MATCHER_BRUTEFORCE_HAMMING)
    keypoints1, descriptors1 = orb.detectAndCompute(im1_scaled, None)
    keypoints2, descriptors2 = orb.detectAndCompute(im2_scaled, None)
    matches = matcher.match(descriptors1, descriptors2, None)
    # Sort matches by score
    matches.sort(key=lambda x: x.distance, reverse=False)
    # Remove not so good matches
    numGoodMatches = int(len(matches) * match_threshold)
    matches = matches[:numGoodMatches]
    distances = [match.distance for match in matches]
    # Get std
    std_distance = np.std(distances)
    
    if plot:
        im3 = cv2.drawMatches(im1_scaled, keypoints1, im2_scaled, keypoints2, matches, None)
        plt.figure(figsize=(20, 10))
        plt.imshow(im3)
        plt.show()

    
    # Extract location of good matches
    points1 = np.zeros((len(matches), 2), dtype=np.float32)
    points2 = np.zeros((len(matches), 2), dtype=np.float32)

    # Get points 
    for i, match in enumerate(matches):
        points1[i, :] = keypoints1[match.queryIdx].pt
        points2[i, :] = keypoints2[match.trainIdx].pt
    # Offset by padding
    points1 -= pad_width
    points2 -= pad_width
    return points1, points2, std_distance

def find_drift_poi(points1, points2):
    """ Find the best fit for the rotation and translation that map points1 to points2 using 
    the RANSAC algorithm (fitting robust to outliers).

    Args:
        points1 (numpy.ndarray, int): list of points (n x 2)
        points2 (numpy.ndarray, int): list of points (n x 2)
    Returns:
        array: x and y translation
    """
    # Find affine transformation matrix from points1 to points2 (i.e. Ax1 = x2)
    h, _ = cv2.estimateAffinePartial2D(points1,points2)
    # Extract x and y translation
    return np.array([h[0,2], h[1,2]])
  
def find_template(reference, check, match_threshold=0.1,window_height=400, window_width=200):
    """ Find template region (fiduciary markers on the mother machine) by cluster of interest points

    Args:
        reference (numpy.ndarray): Reference image from which to extract the template 
        check (numpy.ndarray): Check image to see that the template matches to another timepoint
        match_threshold (float): The fraction of closest matches in Euclidean distance 
        window_height (int): Height of template patch to extract
        window_width (int): Width of template patch to extract
    Returns:
        template (numpy.ndarray): Template match from reference image (window_height x window_width)
        top_left (numpy.ndarray): x and y coordinates of the top-left corner of the template (used to calculate drift) (1 x 2)
    """
    reference_scaled = scale_image(reference)
    check_scaled = scale_image(check)
    # Get interest points
    points1, points2, _ = get_orb_pois(reference_scaled, check_scaled, match_threshold=match_threshold)
    # Get median of interest points in the first image
    median_position = np.median(points1, axis=0)
    # Calculate top left
    top_left = np.array([max(0, int(median_position[0]-window_width/2)), max(0, int(median_position[1]-window_height/2))])
    # Crop image to get template
    template = reference_scaled[top_left[1]:top_left[1]+window_height,top_left[0]:top_left[0]+window_width]
    return template, top_left

def find_drift_template(template, top_left, im2):
    """ Calculate drift using template matching

    Args:
        template(numpy.ndarray): template patch
        top_left (numpy.ndarray): coordinates of top_left of template patch in reference image (1 x 2)
        im2 (numpy.ndarray): image to template match
    Returns:
        array (numpy.ndarray) : drift (1 x 2)
        min_sqdiff: Normalized least square difference between the template match
    """
    im2_scaled = scale_image(im2)
    res = cv2.matchTemplate(im2_scaled, template, cv2.TM_SQDIFF_NORMED)
    min_sqdiff, _, top_left_new, _ = cv2.minMaxLoc(res)
    return top_left_new - top_left, min_sqdiff


import os
import h5py
import torch
import copy
import ipywidgets as ipyw
import scipy
import pandas as pd
import datetime
import time
import itertools
import qgrid
import shutil
import subprocess

from random import shuffle
from torch.autograd import Variable
from torch.utils.data import Dataset, DataLoader
from sklearn.metrics import precision_recall_curve
from scipy.ndimage.interpolation import map_coordinates
from scipy.interpolate import RectBivariateSpline
from scipy import interpolate,ndimage

import skimage as sk
import pickle as pkl
import skimage.morphology
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import numpy as np
from .utils import pandas_hdf5_handler,kymo_handle,writedir
from .trcluster import hdf5lock,dask_controller
from .metrics import object_f_scores

from matplotlib import pyplot as plt

class weightmap_generator:
    def __init__(self,nndatapath,w0,wm_sigma):
        self.nndatapath = nndatapath
        self.w0 = w0
        self.wm_sigma = wm_sigma

    def make_weight_map(self,binary_mask):
        ttl_count = binary_mask.size
        cell_count = np.sum(binary_mask)
        background_count = ttl_count - cell_count
        class_weight = np.array([ttl_count/(background_count+1),ttl_count/(cell_count+1)])
        class_weight = class_weight/np.sum(class_weight)

        labeled = sk.measure.label(binary_mask)
        labels = np.unique(labeled)[1:]

        dist_maps = []
        borders = []
        
        num_labels = len(labels)
        
        if num_labels == 0:
            weight = np.ones(binary_mask.shape)*class_weight[0]
        elif num_labels == 1:
            cell = labeled==1
#             dilated = sk.morphology.binary_dilation(cell)
            eroded = sk.morphology.binary_dilation(cell)
            border = eroded^cell
            weight = np.ones(binary_mask.shape)*class_weight[0]
            weight[binary_mask] += class_weight[1]
#             weight[border] = 0.
        else:
            for i in labels:
                cell = labeled==i
#                 dilated = sk.morphology.binary_dilation(cell)
                eroded = sk.morphology.binary_dilation(cell)
                border = eroded^cell
                borders.append(border)
                dist_map = scipy.ndimage.morphology.distance_transform_edt(~border)
                dist_maps.append(dist_map)
            dist_maps = np.array(dist_maps)
            borders = np.array(borders)
            borders = np.max(borders,axis=0)
            dist_maps = np.sort(dist_maps,axis=0)
            weight = self.w0*np.exp(-((dist_maps[0] + dist_maps[1])**2)/(2*(self.wm_sigma**2)))
            weight[binary_mask] += class_weight[1]
            weight[~binary_mask] += class_weight[0]
#             weight[borders] = 0.
        return weight
    
    def make_weightmaps(self,seg_arr):
        num_indices = seg_arr.shape[0]
        weightmap_arr = []
        for t in range(0,num_indices):
            working_seg_arr = seg_arr[t,0].astype(bool)
            weightmap = self.make_weight_map(working_seg_arr).astype("float32")
            weightmap_arr.append(weightmap)
        weightmap_arr = np.array(weightmap_arr)[:,np.newaxis,:,:]
        return weightmap_arr
            
class data_augmentation:
    def __init__(self,p_flip=0.5,max_rot=10,min_padding=20):
        self.p_flip = p_flip
        self.max_rot = max_rot
        self.min_padding = min_padding
                
#     def make_chunked_kymograph(self,img_arr,chunksize=10):
#         pad = (chunksize - (img_arr.shape[2]%chunksize))*img_arr.shape[1]
#         chunked_arr = np.swapaxes(img_arr,1,2)
#         chunked_arr = chunked_arr.reshape(chunked_arr.shape[0],-1)
#         chunked_arr = np.pad(chunked_arr,((0,0),(0,pad)),'constant',constant_values=0)
#         chunked_arr = chunked_arr.reshape(chunked_arr.shape[0],-1,img_arr.shape[1]*chunksize)
#         chunked_arr = np.swapaxes(chunked_arr,1,2)
#         return chunked_arr
    
    def random_crop(self,img_arr,seg_arr):
        false_arr = np.zeros(img_arr.shape[2:4],dtype=bool)
        random_crop_len_y = np.random.uniform(low=0.1,high=1.,size=(1,img_arr.shape[0]))
        random_crop_len_x = np.random.uniform(low=0.4,high=1.,size=(1,img_arr.shape[0]))
        
        random_crop_len = np.concatenate([random_crop_len_y,random_crop_len_x],axis=0)
        
        random_crop_remainder = 1.-random_crop_len
        random_crop_start = (np.random.uniform(low=0.,high=1.,size=(2,img_arr.shape[0])))*random_crop_remainder
        low_crop = np.floor(random_crop_start*np.array(img_arr.shape[2:4])[:,np.newaxis]).astype('int32')
        high_crop = np.floor(low_crop+(random_crop_len*np.array(img_arr.shape[2:4])[:,np.newaxis])).astype('int32')
#         random_low_samples = np.random.uniform(low=0.,high=0.5,size=(2,img_arr.shape[0]))
#         low_crop = (random_low_samples*np.array(img_arr.shape[2:4])[:,np.newaxis]).astype('int32')
#         remainder = np.array(img_arr.shape[2:4])[:,np.newaxis]-low_crop
#         random_high_samples = np.random.uniform(low=0.5,high=1.,size=(2,img_arr.shape[0]))
#         high_crop = np.floor(random_high_samples*remainder).astype('int32')+low_crop
        out_arr = []
        out_seg_arr = []
        center = (img_arr.shape[2]//2,img_arr.shape[3]//2)
        for t in range(img_arr.shape[0]):
            mask = copy.copy(false_arr)
            working_arr = copy.copy(img_arr[t,0,:,:])
            working_seg_arr = copy.copy(seg_arr[t,0,:,:])

            dim_0_range = (high_crop[0,t] - low_crop[0,t])
            dim_1_range = high_crop[1,t] - low_crop[1,t]
            top_left = (center[0]-dim_0_range//2,center[1]-dim_1_range//2)

            dim_0_maxscale = img_arr.shape[2]/dim_0_range
            dim_1_maxscale = img_arr.shape[3]/dim_1_range

            dim_0_scale = np.clip(np.random.normal(loc=1.0,scale=0.1),0.8,dim_0_maxscale)
            dim_1_scale = np.clip(np.random.normal(loc=1.0,scale=0.1),0.8,dim_1_maxscale)

            rescaled_img = sk.transform.rescale(working_arr[low_crop[0,t]:high_crop[0,t],low_crop[1,t]:high_crop[1,t]],(dim_0_scale,dim_1_scale),preserve_range=True).astype(int)
            rescaled_seg = (sk.transform.rescale(working_seg_arr[low_crop[0,t]:high_crop[0,t],low_crop[1,t]:high_crop[1,t]]==1,(dim_0_scale,dim_1_scale))>0.5).astype("int8")
#             rescaled_border = (sk.transform.rescale(working_seg_arr[low_crop[0,t]:high_crop[0,t],low_crop[1,t]:high_crop[1,t]]==2,(dim_0_scale,dim_1_scale))>0.5)
#             rescaled_seg[rescaled_border] = 2

            top_left = (center[0]-rescaled_img.shape[0]//2,center[1]-rescaled_img.shape[1]//2)        
            working_arr[top_left[0]:top_left[0]+rescaled_img.shape[0],top_left[1]:top_left[1]+rescaled_img.shape[1]] = rescaled_img
            working_seg_arr[top_left[0]:top_left[0]+rescaled_img.shape[0],top_left[1]:top_left[1]+rescaled_img.shape[1]] = rescaled_seg

            mask[top_left[0]:top_left[0]+rescaled_img.shape[0],top_left[1]:top_left[1]+rescaled_img.shape[1]] = True
            working_arr[~mask] = 0
            working_seg_arr[~mask] = False        
            
            out_arr.append(working_arr)
            out_seg_arr.append(working_seg_arr)
        out_arr = np.expand_dims(np.array(out_arr),1)
        out_seg_arr = np.expand_dims(np.array(out_seg_arr),1)
        return out_arr,out_seg_arr
    
    def random_x_flip(self,img_arr,seg_arr,p=0.5):
        choices = np.random.choice(np.array([True,False]),size=img_arr.shape[0],p=np.array([p,1.-p]))
        out_img_arr = copy.copy(img_arr)
        out_seg_arr = copy.copy(seg_arr)
        out_img_arr[choices,0,:,:] = np.flip(img_arr[choices,0,:,:],axis=1)
        out_seg_arr[choices,0,:,:] = np.flip(seg_arr[choices,0,:,:],axis=1)
        return out_img_arr,out_seg_arr
    def random_y_flip(self,img_arr,seg_arr,p=0.5):
        choices = np.random.choice(np.array([True,False]),size=img_arr.shape[0],p=np.array([p,1.-p]))
        out_img_arr = copy.copy(img_arr)
        out_seg_arr = copy.copy(seg_arr)
        out_img_arr[choices,0,:,:] = np.flip(img_arr[choices,0,:,:],axis=2)
        out_seg_arr[choices,0,:,:] = np.flip(seg_arr[choices,0,:,:],axis=2)
        return out_img_arr,out_seg_arr
    
    def change_brightness(self,img_arr,num_control_points=3):
        out_img_arr = copy.copy(img_arr)
        for t in range(img_arr.shape[0]):
            control_points = (np.add.accumulate(np.ones(num_control_points+2))-1.)/(num_control_points+1)
            control_point_locations = (control_points*65535).astype(int)
            orig_locations = copy.copy(control_point_locations)
            random_points = np.random.uniform(low=0,high=65535,size=num_control_points).astype(int)
            sorted_points = np.sort(random_points)
            control_point_locations[1:-1] = sorted_points
            mapping = interpolate.PchipInterpolator(orig_locations, control_point_locations)
            out_img_arr[t,0,:,:] = mapping(img_arr[t,0,:,:])
        return out_img_arr
    
    
    def add_padding(self,img_arr,seg_arr,max_rot=20,min_padding=20):
        hyp_length = np.ceil((img_arr.shape[2]**2+img_arr.shape[3]**2)**(1/2)).astype(int)
        max_rads = ((90-max_rot)/360)*(2*np.pi)
        min_rads = (90/360)*(2*np.pi)
        max_y = np.maximum(np.ceil(hyp_length*np.sin(max_rads)),np.ceil(hyp_length*np.sin(min_rads))).astype(int)
        max_x = np.maximum(np.ceil(hyp_length*np.cos(max_rads)),np.ceil(hyp_length*np.cos(min_rads))).astype(int)
        delta_y = max_y-img_arr.shape[2]
        delta_x = max_x-img_arr.shape[3]
        if delta_x % 2 == 1:
            delta_x+=1
        if delta_y % 2 == 1:
            delta_y+=1
        delta_y = np.maximum(delta_y,2*min_padding)
        delta_x = np.maximum(delta_x,2*min_padding)
        padded_img_arr = np.pad(img_arr, ((0,0),(0,0),(delta_y//2,delta_y//2),(delta_x//2,delta_x//2)), 'constant', constant_values=0)
        padded_seg_arr = np.pad(seg_arr, ((0,0),(0,0),(delta_y//2,delta_y//2),(delta_x//2,delta_x//2)), 'constant', constant_values=0)
        return padded_img_arr,padded_seg_arr
    
    def translate(self,pad_img_arr,pad_seg_arr,img_arr,seg_arr):
        trans_img_arr = copy.copy(pad_img_arr)
        trans_seg_arr = copy.copy(pad_seg_arr)
        delta_y = pad_img_arr.shape[2] - img_arr.shape[2]
        delta_x = pad_img_arr.shape[3] - img_arr.shape[3]
        for t in range(pad_img_arr.shape[0]):
            trans_y = np.random.randint(-(delta_y//2),high=delta_y//2)
            trans_x = np.random.randint(-(delta_x//2),high=delta_x//2)
            trans_img_arr[t,0,delta_y//2:delta_y//2+img_arr.shape[2],delta_x//2:delta_x//2+img_arr.shape[3]] = 0
            trans_seg_arr[t,0,delta_y//2:delta_y//2+img_arr.shape[2],delta_x//2:delta_x//2+img_arr.shape[3]] = 0
            trans_img_arr[t,0,delta_y//2+trans_y:delta_y//2+img_arr.shape[2]+trans_y,delta_x//2+trans_x:delta_x//2+img_arr.shape[3]+trans_x] =\
            pad_img_arr[t,0,delta_y//2:delta_y//2+img_arr.shape[2],delta_x//2:delta_x//2+img_arr.shape[3]]
            trans_seg_arr[t,0,delta_y//2+trans_y:delta_y//2+img_arr.shape[2]+trans_y,delta_x//2+trans_x:delta_x//2+img_arr.shape[3]+trans_x] =\
            pad_seg_arr[t,0,delta_y//2:delta_y//2+img_arr.shape[2],delta_x//2:delta_x//2+img_arr.shape[3]]
        return trans_img_arr,trans_seg_arr
    
    def rotate(self,img_arr,seg_arr,max_rot=20):
        rot_img_arr = copy.copy(img_arr)
        rot_seg_arr = copy.copy(seg_arr)
        for t in range(img_arr.shape[0]):
            r = np.random.uniform(low=-max_rot,high=max_rot)
            rot_img_arr[t,0,:,:] = sk.transform.rotate(img_arr[t,0,:,:],r,preserve_range=True).astype("int32")
            rot_seg = (sk.transform.rotate(seg_arr[t,0,:,:]==1,r)>0.5).astype("int8")
#             rot_border = sk.transform.rotate(seg_arr[t,:,:]==2,r)>0.5
#             rot_seg[rot_border] = 2
            rot_seg_arr[t,0,:,:] = rot_seg
        return rot_img_arr,rot_seg_arr

    def deform_img_arr(self,img_arr,seg_arr):
        def_img_arr = copy.copy(img_arr)
        def_seg_arr = copy.copy(seg_arr)
        for t in range(img_arr.shape[0]):
            y_steps = np.linspace(0.,4.,num=img_arr.shape[2])
            x_steps = np.linspace(0.,4.,num=img_arr.shape[3])
            grid = np.random.normal(scale=1.,size=(2,4,4))
            dx = RectBivariateSpline(np.arange(4),np.arange(4),grid[0]).ev(y_steps[:,np.newaxis],x_steps[np.newaxis,:])
            dy = RectBivariateSpline(np.arange(4),np.arange(4),grid[1]).ev(y_steps[:,np.newaxis],x_steps[np.newaxis,:])
            y,x = np.meshgrid(np.arange(img_arr.shape[2]), np.arange(img_arr.shape[3]), indexing='ij')
            indices = np.reshape(y+dy, (-1, 1)), np.reshape(x+dx, (-1, 1))
            elastic_img = map_coordinates(img_arr[t,0,:,:], indices, order=1).reshape(img_arr.shape[2:4])
            
            def_img_arr[t,0,:,:] = elastic_img
            
            elastic_cell = (map_coordinates(seg_arr[t,0,:,:]==1, indices, order=1).reshape(seg_arr.shape[2:4])>0.5)
            elastic_cell = sk.morphology.binary_closing(elastic_cell)
#             elastic_border = (map_coordinates(seg_arr[t,:,:]==2, indices, order=1).reshape(seg_arr.shape[1:3])>0.5)
            def_seg_arr[t,0,elastic_cell] = 1
#             def_seg_arr[t,elastic_border] = 2
        return def_img_arr,def_seg_arr

    
    def get_augmented_data(self,img_arr,seg_arr):
        img_arr,seg_arr = self.random_crop(img_arr,seg_arr)
        img_arr,seg_arr = self.random_x_flip(img_arr,seg_arr,p=self.p_flip)
        img_arr,seg_arr = self.random_y_flip(img_arr,seg_arr,p=self.p_flip)
        img_arr = self.change_brightness(img_arr)
        pad_img_arr,pad_seg_arr = self.add_padding(img_arr,seg_arr,max_rot=self.max_rot+5)
        img_arr,seg_arr = self.translate(pad_img_arr,pad_seg_arr,img_arr,seg_arr)
        del pad_img_arr
        del pad_seg_arr
        img_arr,seg_arr = self.rotate(img_arr,seg_arr,max_rot=self.max_rot)
        img_arr,seg_arr = self.deform_img_arr(img_arr,seg_arr)
        img_arr,seg_arr = (img_arr.astype("int32"),seg_arr.astype("int8"))
        return img_arr,seg_arr

class UNet_Training_DataLoader:
    def __init__(self,nndatapath="",experimentname="",trainpath="",testpath="",valpath="",augment=False):
        self.nndatapath = nndatapath
        self.experimentname = experimentname
        self.trainpath = trainpath
        self.testpath = testpath
        self.valpath = valpath
        
        self.trainname = self.trainpath.split("/")[-1]
        self.testname = self.testpath.split("/")[-1]
        self.valname = self.valpath.split("/")[-1]
        
        self.metapath = self.nndatapath + "/metadata.hdf5"
        
    def get_metadata(self,headpath):
        meta_handle = pandas_hdf5_handler(headpath + "/metadata.hdf5")
        global_handle = meta_handle.read_df("global",read_metadata=True)
        kymo_handle = meta_handle.read_df("kymograph",read_metadata=True)
        fovdf = kymo_handle.reset_index(inplace=False)
        fovdf = fovdf.set_index(["fov","row","trench"], drop=True, append=False, inplace=False)
        fovdf = fovdf.sort_index()
        
        channel_list = global_handle.metadata["channels"]
        fov_list = kymo_handle['fov'].unique().tolist()
        t_len = len(kymo_handle.index.get_level_values("timepoints").unique())
        trench_dict = {fov:len(fovdf.loc[fov]["trenchid"].unique()) for fov in fov_list}
        shape_y = kymo_handle.metadata["kymograph_params"]["ttl_len_y"]
        shape_x = kymo_handle.metadata["kymograph_params"]["trench_width_x"]
        kymograph_img_shape = tuple((shape_y,shape_x))
        return channel_list,fov_list,t_len,trench_dict,kymograph_img_shape
    
    def get_selection(self,channel,trench_dict,fov_list,t_subsample_step,t_range,max_trenches,kymograph_img_shape,selectionname):        
        fov_list = list(fov_list)
        ttl_trench_count = np.sum(np.array([trench_dict[fov] for fov in fov_list]))
        ttl_trench_count = min(ttl_trench_count,max_trenches)
        num_t = len(range(t_range[0],t_range[1]+1,t_subsample_step))
        ttl_imgs = ttl_trench_count*num_t
        print("Total Number of Trenches: " + str(ttl_trench_count))
        print("Total Number of Timepoints: " + str(num_t))
        print("Total Number of Images: " + str(ttl_imgs))
        selection = tuple((channel,fov_list,t_subsample_step,t_range,max_trenches,ttl_imgs,kymograph_img_shape))
        setattr(self, selectionname + "_selection", selection)
        
    def inter_get_selection(self,headpath,selectionname):
        channel_list,fov_list,t_len,trench_dict,kymograph_img_shape = self.get_metadata(headpath)
        selection = ipyw.interactive(self.get_selection, {"manual":True}, channel=ipyw.Dropdown(options=channel_list,value=channel_list[0],description='Feature Channel:',disabled=False),\
                trench_dict=ipyw.fixed(trench_dict),fov_list=ipyw.SelectMultiple(options=fov_list),\
                t_subsample_step=ipyw.IntSlider(value=1, min=1, max=50, step=1),\
                t_range=ipyw.IntRangeSlider(value=[0, t_len-1],min=0,max=t_len-1,step=1,disabled=False,continuous_update=False),\
                max_trenches=ipyw.IntText(value=1,description='Maximum Trenches per FOV: ',disabled=False),\
                kymograph_img_shape=ipyw.fixed(kymograph_img_shape),\
                selectionname=ipyw.fixed(selectionname));
        display(selection)
    
    def export_chunk(self,selectionname,file_idx,augment,file_trench_indices,weight_grid_list):
        selection = getattr(self,selectionname + "_selection")
        datapath = getattr(self,selectionname + "path")
        dataname = getattr(self,selectionname + "name")
        
        img_path = datapath + "/kymograph/kymograph_" + str(file_idx) + ".hdf5"
        seg_path = datapath + "/fluorsegmentation/segmentation_" + str(file_idx) + ".hdf5"
        nndatapath = self.nndatapath + "/" + selectionname + "_" + str(file_idx) + ".hdf5"
        
        with h5py.File(img_path,"r") as imgfile:
            img_arr = imgfile[selection[0]][file_trench_indices,selection[3][0]:selection[3][1]+1:selection[2]]
            img_arr = img_arr.reshape(img_arr.shape[0]*img_arr.shape[1],img_arr.shape[2],img_arr.shape[3])
            img_arr = img_arr[:,np.newaxis,:,:]
            img_arr = img_arr.astype('int32')

        with h5py.File(seg_path,"r") as segfile:
            seg_arr = segfile["data"][file_trench_indices,selection[3][0]:selection[3][1]+1:selection[2]]
            seg_arr = seg_arr.reshape(seg_arr.shape[0]*seg_arr.shape[1],seg_arr.shape[2],seg_arr.shape[3])
            seg_arr = seg_arr[:,np.newaxis,:,:]
            seg_arr = seg_arr.astype('int8')

        if augment:
            img_arr,seg_arr = self.data_augmentation.get_augmented_data(img_arr,seg_arr)
            
        chunk_shape = (1,1,img_arr.shape[2],img_arr.shape[3])
            
        with h5py.File(nndatapath,"w") as outfile:
            img_handle = outfile.create_dataset("img",data=img_arr,chunks=chunk_shape,dtype='int32')
            seg_handle = outfile.create_dataset("seg",data=seg_arr,chunks=chunk_shape,dtype='int8')
            
        for item in weight_grid_list:
            w0,wm_sigma = item
            weightmap_gen = weightmap_generator(self.nndatapath,w0,wm_sigma)
            weightmap_arr = weightmap_gen.make_weightmaps(seg_arr)
            with h5py.File(nndatapath,"a") as outfile:
                weightmap_handle = outfile.create_dataset("weight_" + str(item),data=weightmap_arr,chunks=chunk_shape,dtype='int32')
            
        return file_idx
            
    def gather_chunks(self,outputdf,output_metadata,selectionname,file_idx_list,weight_grid_list):
        nnoutputpath = self.nndatapath + "/" + selectionname + ".hdf5"
                
        tempdatapath = self.nndatapath + "/" + selectionname + "_" + str(file_idx_list[0]) + ".hdf5"
        with h5py.File(tempdatapath,"r") as infile:
            output_shape = (len(outputdf.index),1,infile["img"].shape[2],infile["img"].shape[3])
            chunk_shape = (1,1,infile["img"].shape[2],infile["img"].shape[3])
        
        with h5py.File(nnoutputpath,"w") as outfile:
            img_handle = outfile.create_dataset("img",output_shape,chunks=chunk_shape,dtype='int32')
            seg_handle = outfile.create_dataset("seg",output_shape,chunks=chunk_shape,dtype='int8')
            for item in weight_grid_list:
                weightmap_handle = outfile.create_dataset("weight_" + str(item),output_shape,chunks=chunk_shape,dtype='int32')
        
        current_idx = 0
        for file_idx in file_idx_list:
            nndatapath = self.nndatapath + "/" + selectionname + "_" + str(file_idx) + ".hdf5"
            with h5py.File(nndatapath,"r") as infile:
                img_arr = infile["img"][:]
                seg_arr = infile["seg"][:]
                weight_arr_list = []
                for item in weight_grid_list:
                    weight_arr_list.append(infile["weight_" + str(item)][:])
            num_indices = img_arr.shape[0]
            with h5py.File(nnoutputpath,"a") as outfile:
                outfile["img"][current_idx:current_idx+num_indices] = img_arr
                outfile["seg"][current_idx:current_idx+num_indices] = seg_arr
                for i,item in enumerate(weight_grid_list):
                    outfile["weight_" + str(item)][current_idx:current_idx+num_indices] = weight_arr_list[i]
            current_idx += num_indices
            os.remove(nndatapath)

    def export_data(self,selectionname,dask_controller,weight_grid_list,augment=False):
        
        dask_controller.futures = {}
        
        selection = getattr(self,selectionname + "_selection")
        datapath = getattr(self,selectionname + "path")
        dataname = getattr(self,selectionname + "name")
        
        input_meta_handle = pandas_hdf5_handler(datapath + "/metadata.hdf5")
        output_meta_handle = pandas_hdf5_handler(self.metapath)
        
        trenchdf_list = []
        
        kymodf = input_meta_handle.read_df("kymograph",read_metadata=True)
        fovdf = kymodf.reset_index(inplace=False)
        fovdf = fovdf.set_index(["fov","row","trench"], drop=True, append=False, inplace=False)
        fovdf = fovdf.sort_index()
        
        trenchdf = fovdf.loc[selection[1]]
        trenchdf = trenchdf.reset_index(inplace=False)
        trenchdf = trenchdf.set_index(["trenchid","timepoints"], drop=True, append=False, inplace=False)
        trenchdf = trenchdf.sort_index()
        
        trenches = trenchdf.index.get_level_values("trenchid").unique().tolist()
        shuffle(trenches)
        trenches = np.sort(trenches[:selection[4]])
        
        filedf = trenchdf.loc[pd.IndexSlice[trenches, selection[3][0]:selection[3][1]+1:selection[2]], :]        
        filedf = filedf.reset_index(inplace=False)
        filedf = filedf.set_index(["File Index","File Trench Index"], drop=True, append=False, inplace=False)
        filedf = filedf.sort_index()
                
        filelist = filedf.index.get_level_values("File Index").unique().tolist()
        for file_idx in filelist:
            file_trenchdf = filedf.loc[file_idx]
            file_trench_indices = file_trenchdf.index.get_level_values("File Trench Index").unique().tolist()
            future = dask_controller.daskclient.submit(self.export_chunk,selectionname,file_idx,augment,file_trench_indices,weight_grid_list,retries=1)
            dask_controller.futures["File Number: " + str(file_idx)] = future
                
        outputdf = filedf.reset_index(inplace=False)
        outputdf = outputdf.set_index(["trenchid","timepoints"], drop=True, append=False, inplace=False)
        outputdf = outputdf.sort_index()
        
        del outputdf["File Index"]
        del outputdf["File Trench Index"]
        
        selection_keys = ["channel", "fov_list", "t_subsample_step", "t_range", "max_trenches", "ttl_imgs", "kymograph_img_shape"]
        selection = {selection_keys[i]:item for i,item in enumerate(selection)}
        selection["experiment_name"],selection["data_name"] = (self.experimentname, dataname)
        selection["W0 List"], selection["Wm Sigma List"] = (self.grid_dict['W0 (Border Region Weight):'],self.grid_dict['Wm Sigma (Border Region Spread):'])
                          
        output_metadata = {"nndataset" : selection}
        
        segparampath = datapath + "/fluorescent_segmentation.par"
        with open(segparampath, 'rb') as infile:
            seg_param_dict = pkl.load(infile)
        
        output_metadata["segmentation"] = seg_param_dict
        
        input_meta_handle = pandas_hdf5_handler(datapath + "/metadata.hdf5")
        for item in ["global","kymograph"]:
            indf = input_meta_handle.read_df(item,read_metadata=True)
            output_metadata[item] = indf.metadata
        
        output_meta_handle.write_df(selectionname,outputdf,metadata=output_metadata)
        
        file_idx_list = dask_controller.daskclient.gather([dask_controller.futures["File Number: " + str(file_idx)] for file_idx in filelist])
        self.gather_chunks(outputdf,output_metadata,selectionname,file_idx_list,weight_grid_list)
        
        
    def display_grid(self):
        tab_dict = {'W0 (Border Region Weight):':[1., 3., 5., 10.],'Wm Sigma (Border Region Spread):':[1., 2., 3., 4., 5.]}
        children = [ipyw.SelectMultiple(options=val,value=(val[1],),description=key,disabled=False) for key,val in tab_dict.items()]
        self.tab = ipyw.Tab()
        self.tab.children = children
        for i,key in enumerate(tab_dict.keys()):
            self.tab.set_title(i, key[:-1])
        return self.tab
    
    def get_grid_params(self):
        if hasattr(self,'tab'):
            self.grid_dict = {child.description:child.value for child in self.tab.children}
            delattr(self, 'tab')
        elif hasattr(self,'grid_dict'):
            pass
        else:
            raise "No selection defined."
        print("======== Grid Params ========")
        for key,val in self.grid_dict.items():
            print(key + " " + str(val))
        
    def export_all_data(self,n_workers=20,memory='4GB'):
        writedir(self.nndatapath,overwrite=True)
        
        grid_keys = self.grid_dict.keys()
        grid_combinations = list(itertools.product(*list(self.grid_dict.values())))

        self.data_augmentation = data_augmentation()
        
        dask_cont = dask_controller(walltime='01:00:00',local=False,n_workers=n_workers,memory=memory)
        dask_cont.startdask()
#         dask_cont.daskcluster.start_workers()
        dask_cont.displaydashboard()
        
        try:        
            for selectionname in ["train","test","val"]:
                if selectionname == "train":
                    self.export_data(selectionname,dask_cont,grid_combinations,augment=True)
                else:
                    self.export_data(selectionname,dask_cont,grid_combinations,augment=False)
            dask_cont.shutdown()
        except:
            dask_cont.shutdown()
            raise
                        
class GridSearch:
    def __init__(self,nndatapath,numepochs=50):
        self.nndatapath = nndatapath
        self.numepochs = numepochs
        
    def display_grid(self):
        meta_handle = pandas_hdf5_handler(self.nndatapath + "/metadata.hdf5")
        trainmeta = meta_handle.read_df("train",read_metadata=True).metadata["nndataset"]
        w0_list,wm_sigma_list = trainmeta["W0 List"],trainmeta["Wm Sigma List"]
        
        self.tab_dict = {'Batch Size:':[5, 10, 25],'Layers:':[2, 3, 4],\
           'Hidden Size:':[16, 32, 64],'Learning Rate:':[0.001, 0.005, 0.01, 0.05],\
           'Momentum:':[0.9, 0.95, 0.99],'Weight Decay:':[0.0001,0.0005, 0.001],\
           'Dropout:':[0., 0.3, 0.5, 0.7], 'w0:':w0_list, 'wm sigma':wm_sigma_list}
                  
        children = [ipyw.SelectMultiple(options=val,value=(val[1],),description=key,disabled=False) for key,val in self.tab_dict.items()]
        self.tab = ipyw.Tab()
        self.tab.children = children
        for i,key in enumerate(self.tab_dict.keys()):
            self.tab.set_title(i, key[:-1])
        return self.tab
    
    def get_grid_params(self):
        self.grid_dict = {child.description:child.value for child in self.tab.children}
        print("======== Grid Params ========")
        for key,val in self.grid_dict.items():
            print(key + " " + str(val))
            
    def generate_pyscript(self,run_idx,grid_params):
        import_line = "import trenchripper as tr"
        trainer_line = "nntrainer = tr.unet.UNet_Trainer(\"" + self.nndatapath + "\"," + str(run_idx) + \
        ",gpuon=True,numepochs=" + str(self.numepochs) + ",batch_size=" + str(grid_params[0])+",layers=" + \
        str(grid_params[1])+",hidden_size=" + str(grid_params[2]) + ",lr=" + str(grid_params[3]) + \
        ",momentum=" + str(grid_params[4]) + ",weight_decay=" + str(grid_params[5])+",dropout="+str(grid_params[6]) + \
        ",w0=" + str(grid_params[7]) + ",wm_sigma=" + str(grid_params[8]) + ")"
        train_line = "nntrainer.train_model()"
        pyscript = "\n".join([import_line,trainer_line,train_line])
        with open(self.nndatapath + "/models/scripts/" + str(run_idx) + ".py", "w") as scriptfile:
            scriptfile.write(pyscript)
    
    def generate_sbatchscript(self,run_idx,hours,cores,mem,gres):
        shebang = "#!/bin/bash"
        core_line = "#SBATCH -c " + str(cores)
        hour_line = "#SBATCH -t " + str(hours) + ":00:00"
        gpu_lines = "#SBATCH -p gpu\n#SBATCH --gres=" + gres
        mem_line = "#SBATCH --mem=" + mem
        report_lines = "#SBATCH -o " + self.nndatapath + "/models/scripts/" + str(run_idx) +\
        ".out\n#SBATCH -e " + self.nndatapath + "/models/scripts/" + str(run_idx) + ".err\n"
        
        run_line = "python -u " + self.nndatapath + "/models/scripts/" + str(run_idx) + ".py"
        
        sbatchscript = "\n".join([shebang,core_line,hour_line,gpu_lines,mem_line,report_lines,run_line])
        with open(self.nndatapath + "/models/scripts/" + str(run_idx) + ".sh", "w") as scriptfile:
            scriptfile.write(sbatchscript)
    
    def run_sbatchscript(self,run_idx):
        cmd = ["sbatch",self.nndatapath + "/models/scripts/" + str(run_idx) + ".sh"]
        subprocess.run(cmd)

    def run_grid_search(self,hours=12,cores=2,mem="8G",gres="gpu:1"):
        
        grid_keys = self.grid_dict.keys()
        grid_combinations = list(itertools.product(*list(self.grid_dict.values())))
        writedir(self.nndatapath + "/models",overwrite=True)
        writedir(self.nndatapath + "/models/scripts",overwrite=True)
        
        self.run_indices = []
        
        for run_idx,grid_params in enumerate(grid_combinations):
            self.generate_pyscript(run_idx,grid_params)
            self.generate_sbatchscript(run_idx,hours,cores,mem,gres)
            self.run_sbatchscript(run_idx)
            self.run_indices.append(run_idx)
            
    def cancel_all_runs(self,username):
        for run_idx in self.run_indices:
            cmd = ["scancel","-p","gpu","--user=" + username]
            subprocess.Popen(cmd,shell=True,stdin=None,stdout=None,stderr=None,close_fds=True)
            

class SegmentationDataset(Dataset):
    def __init__(self,filepath,weightchannel="",training=False):
        self.filepath = filepath
        self.weightchannel = weightchannel
        self.training = training
        self.chunksize = 1000
        with h5py.File(self.filepath,"r") as infile:
            self.shape = infile["img"].shape
        self.current_chunk = 0
        self.load_chunk(self.current_chunk)
    def load_chunk(self,chunk_idx):
        with h5py.File(self.filepath,"r") as infile:
            self.img_data = infile["img"][chunk_idx*self.chunksize:(chunk_idx+1)*self.chunksize]
            if self.training:
                self.seg_data = infile["seg"][chunk_idx*self.chunksize:(chunk_idx+1)*self.chunksize]
                self.weight_data = infile[self.weightchannel][chunk_idx*self.chunksize:(chunk_idx+1)*self.chunksize]
        self.current_chunk = chunk_idx
    def __len__(self):
        with h5py.File(self.filepath,"r") as infile:
            out_len = infile["img"].shape[0]
        return out_len
    def __getitem__(self,idx):
        idx_chunk = idx//self.chunksize
        subidx = idx%self.chunksize
        if idx_chunk != self.current_chunk:
            self.load_chunk(idx_chunk)
        if self.training:
            sample = {'img': self.img_data[subidx], 'seg': self.seg_data[subidx], self.weightchannel: self.weight_data[subidx]}
        else:
            sample = {'img': self.img_data[subidx]}
        return sample

class double_conv(nn.Module):
    '''(Conv => BatchNorm =>ReLU) twice'''
    def __init__(self,in_ch,out_ch):
        super().__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, 3, padding=1),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_ch, out_ch, 3, padding=1),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True)
        )
        
    def forward(self,x):
        x = self.conv(x)
        return x

class inconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super().__init__()
        self.conv = double_conv(in_ch, out_ch)

    def forward(self, x):
        x = self.conv(x)
        return x
    
class down(nn.Module):
    def __init__(self, in_ch, out_ch):
        super().__init__()
        self.downconv = nn.Sequential(
            nn.MaxPool2d(2),
            double_conv(in_ch, out_ch))

    def forward(self, x):
        x = self.downconv(x)
        return x
    
class up(nn.Module):
    def __init__(self, in_ch, out_ch, bilinear=False):
        super().__init__()

        #  would be a nice idea if the upsampling could be learned too,
        #  but my machine do not have enough memory to handle all those weights
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
        else:
            self.up = nn.ConvTranspose2d(in_ch//2, in_ch//2, 2, stride=2)
        self.conv = double_conv(in_ch, out_ch)
        
    def forward(self, x1, x2):
        x1 = self.up(x1)
        
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, (diffX // 2, diffX - diffX//2,
                        diffY // 2, diffY - diffY//2))
        
        # for padding issues, see 
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd

        x = torch.cat([x2, x1], dim=1)
        x = self.conv(x)
        return x
    
class outconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super().__init__()
        self.conv = nn.Conv2d(in_ch, out_ch, 1)

    def forward(self, x):
        x = self.conv(x)
        return x
    
class UNet(nn.Module):
    def __init__(self,n_channels,n_classes,layers=3,hidden_size=64,dropout=0.,withsoftmax=False):
        super().__init__()
        self.inc = inconv(n_channels, hidden_size)
        self.downlist = nn.ModuleList([down(hidden_size*(2**i), hidden_size*(2**(i+1))) for i in range(0,layers-1)] + [down(hidden_size*(2**(layers-1)), hidden_size*(2**(layers-1)))])
        self.uplist = nn.ModuleList([up(hidden_size*(2**i), hidden_size*(2**(i-2))) for i in reversed(range(2,layers+1))] + [up(hidden_size*2, hidden_size)])
        self.outc = outconv(hidden_size, n_classes)
        self.drop = nn.Dropout(p=dropout)
        self.withsoftmax = withsoftmax
    def uniforminit(self):
        for param in self.named_parameters():
            param[1].data.uniform_(-0.05,0.05)
    def forward(self, x):
        xlist = [self.inc(x)]
        for item in self.downlist:
            xlist.append(item(xlist[-1]))
        x = xlist[-1]
        x = self.drop(x)
        for i,item in enumerate(self.uplist):
            x = item(x, xlist[-(i+2)])
        x = self.outc(x)
        if self.withsoftmax:
            x = F.softmax(x,dim=1)
        return x
    
class UNet_Trainer:
    
    def __init__(self,nndatapath,model_number,numepochs=10,batch_size=100,layers=3,hidden_size=64,lr=0.005,momentum=0.95,weight_decay=0.0005,dropout=0.,\
                 w0=5.,wm_sigma=3.,gpuon=False):
        self.nndatapath = nndatapath
        self.model_number = model_number
        
        self.numepochs = numepochs
        self.batch_size = batch_size
        self.gpuon = gpuon
        
        self.layers = layers
        self.hidden_size = hidden_size
        self.dropout = dropout
        self.lr = lr
        self.momentum = momentum
        self.weight_decay = weight_decay
        self.w0 = w0
        self.wm_sigma = wm_sigma
        
        self.model = UNet(1,2,layers=layers,hidden_size=hidden_size,dropout=dropout,withsoftmax=True)
        self.model.uniforminit()
        if gpuon:
            self.model = self.model.cuda()
            
        self.optimizer = optim.SGD(self.model.parameters(), lr = self.lr,momentum=self.momentum,weight_decay=self.weight_decay)
                
    def removefile(self,path):
        if os.path.exists(path):
            os.remove(path)
        
    def load_model(self,paramspath):
        if self.gpuon:
            device = torch.device("cuda")
            self.model.load_state_dict(torch.load(paramspath))
        else:
            device = torch.device('cpu')
            self.model.load_state_dict(torch.load(paramspath, map_location=device))
    
    def train(self,x,y,weightmaps):
        self.optimizer.zero_grad()
        fx = self.model.forward(x)
        fx = torch.log(fx)
        
        nll = F.nll_loss(fx,y,reduction='none')*weightmaps
        
        mean_nll = torch.mean(nll)
        mean_nll.backward()
        self.optimizer.step()
        
        nll = torch.sum(nll)
        return nll

    def test(self,x,y,weightmaps):
        fx = self.model.forward(x)
        fx = torch.log(fx)
        
        nll = F.nll_loss(fx,y,reduction='none')*weightmaps
        
        nll = torch.sum(nll)
        return nll
    
    def perepoch(self,e,train_iter,test_iter,val_iter,train_data_shape,test_data_shape,val_data_shape):
        
        now = datetime.datetime.now()
        
        print('=======epoch ' + str(e) + '=======')
        self.model.train()
        total_train_nll = 0.
        num_train_batches = len(train_iter)
        for i,b in enumerate(train_iter):
            img_arr,seg_arr,weightmaps = (b['img'].numpy(),b['seg'].numpy(),b['weight_' + str(tuple([self.w0,self.wm_sigma]))].numpy())
            
            seg_arr,weightmaps = seg_arr[:,0],weightmaps[:,0]
            x = torch.Tensor(img_arr)
            y = torch.LongTensor(seg_arr)
            weightmaps = torch.Tensor(weightmaps)
            if self.gpuon:
                x = x.cuda()
                y = y.cuda()
                weightmaps = weightmaps.cuda()
#                 weights = weights.cuda()
            nll = self.train(x,y,weightmaps)
            total_train_nll += nll.detach().cpu().numpy()
#             if (i%100 == 0) and self.saveparams:
#                 torch.save(self.model.state_dict(), self.nnpath + "/model_layers=" + str(self.layers) + "_hidden_size=" + str(self.hidden_size) +\
#                            "_dropout=" + str(self.dropout) + '_lr=' + str(self.lr) + '_momentum=' + str(self.momentum) + "_epoch_" + str(e) + "_step_" + str(i) +".pt")
            del x
            del y
            del weightmaps
            del nll
            torch.cuda.empty_cache()
        avgtrainnll = total_train_nll/(np.prod(np.array(train_data_shape)))
        print('Mean Train NLL: ' + str(avgtrainnll))
        self.model.eval()
        total_val_nll = 0.
        for i,b in enumerate(val_iter):
            img_arr,seg_arr,weightmaps = (b['img'].numpy(),b['seg'].numpy(),b['weight_' + str(tuple([self.w0,self.wm_sigma]))].numpy())
            seg_arr,weightmaps = seg_arr[:,0],weightmaps[:,0]
            x = torch.Tensor(img_arr)
            y = torch.LongTensor(seg_arr)
            weightmaps = torch.Tensor(weightmaps)
            if self.gpuon:
                x = x.cuda()
                y = y.cuda()
                weightmaps = weightmaps.cuda()
#                 weights = weights.cuda()
            nll = self.test(x,y,weightmaps)
            total_val_nll += nll.detach().cpu().numpy()
            del x
            del y
            del weightmaps
            del nll
            torch.cuda.empty_cache()
        avgvalnll = total_val_nll/(np.prod(np.array(val_data_shape)))
        print('Mean Val NLL: ' + str(avgvalnll))
        
        total_test_nll = 0.
        for i,b in enumerate(test_iter):
            img_arr,seg_arr,weightmaps = (b['img'].numpy(),b['seg'].numpy(),b['weight_' + str(tuple([self.w0,self.wm_sigma]))].numpy())
            seg_arr,weightmaps = seg_arr[:,0],weightmaps[:,0]
            x = torch.Tensor(img_arr)
            y = torch.LongTensor(seg_arr)
            weightmaps = torch.Tensor(weightmaps)
            if self.gpuon:
                x = x.cuda()
                y = y.cuda()
                weightmaps = weightmaps.cuda()
#                 weights = weights.cuda()
            nll = self.test(x,y,weightmaps)
            total_test_nll += nll.detach().cpu().numpy()
            del x
            del y
            del weightmaps
            del nll
            torch.cuda.empty_cache()
        avgtestnll = total_test_nll/(np.prod(np.array(test_data_shape)))
        print('Mean Test NLL: ' + str(avgtestnll))
        
        entry = [[self.model_number,self.batch_size,self.layers,self.hidden_size,self.lr,self.momentum,self.weight_decay,\
                  self.dropout,self.w0,self.wm_sigma,e,avgtrainnll,avgvalnll,avgtestnll,str(now)]]

        df_out = pd.DataFrame(data=entry,columns=['Model #','Batch Size','Layers','Hidden Size','Learning Rate','Momentum','Weight Decay',\
                            'Dropout',"W0 Weight","Wm Sigma",'Epoch','Train Loss','Val Loss','Test Loss','Date/Time'])
        df_out = df_out.set_index(['Model #','Epoch'], drop=True, append=False, inplace=False)
        df_out = df_out.sort_index()
        
        return df_out
        
    def write_metadata(self,filepath,iomode,df_out):
        meta_handle = pandas_hdf5_handler(filepath)
        if os.path.exists(filepath):
            ind = df_out.index[0]
            df_in = meta_handle.read_df("data")
            df_mask = ~df_in.index.isin([ind])
            df_in = df_in[df_mask]
            df_out = pd.concat([df_in, df_out])
        meta_handle.write_df("data",df_out)
        
    def get_fscore(self,iterator,data_shape):
        y_true = []
        y_scores = []
        for i,b in enumerate(iterator):
            img_arr,y = (b['img'].numpy(),b['seg'].numpy())
            x = torch.Tensor(img_arr)
            if self.gpuon:
                x = x.cuda()
            fx = self.model.forward(x).detach().cpu().numpy()
#             y_true.append(y.flatten())
#             y_scores.append(fx[:,1].flatten())
            
            y_true.append(y[:,0])
            y_scores.append(fx[:,1])
            
            del x
            del y
            torch.cuda.empty_cache()
            
        y_true = np.concatenate(y_true,axis=0)
        y_scores = np.concatenate(y_scores,axis=0)
        precisions, recalls, thresholds = precision_recall_curve(y_true.flatten(), y_scores.flatten())        
        fscores = 2*((precisions*recalls)/(precisions+recalls))
        best_idx = np.nanargmax(fscores)
        precision, recall, fscore, threshold = (precisions[best_idx], recalls[best_idx], fscores[best_idx], thresholds[best_idx])
        
        y_true = y_true.astype(bool)
        y_pred = y_scores>threshold
        
        all_f_scores = []
        for i in range(y_true.shape[0]):
            _,_,f_score = object_f_scores(y_true[i],y_pred[i])
            all_f_scores += f_score.tolist()
        all_f_scores = np.array(all_f_scores)
        all_f_scores = all_f_scores[~np.isnan(all_f_scores)]
        
        return precision, recall, fscore, threshold, all_f_scores
    
    def train_model(self):
        timestamp = datetime.datetime.now()
        start = time.time()
        writedir(self.nndatapath + "/models", overwrite=False)
        self.removefile(self.nndatapath + "/models/training_metadata_" + str(self.model_number) + ".hdf5")
        
        train_data = SegmentationDataset(self.nndatapath + "/train.hdf5",weightchannel='weight_' + str(tuple([self.w0,self.wm_sigma])),training=True)
        test_data = SegmentationDataset(self.nndatapath + "/test.hdf5",weightchannel='weight_' + str(tuple([self.w0,self.wm_sigma])),training=True)
        val_data = SegmentationDataset(self.nndatapath + "/val.hdf5",weightchannel='weight_' + str(tuple([self.w0,self.wm_sigma])),training=True)
        
        train_data_shape = train_data.shape
        test_data_shape = test_data.shape
        val_data_shape = val_data.shape
                
        for e in range(0,self.numepochs):
            train_iter = DataLoader(train_data,batch_size=self.batch_size,shuffle=True)
            test_iter = DataLoader(test_data,batch_size=self.batch_size,shuffle=True)
            val_iter = DataLoader(val_data,batch_size=self.batch_size,shuffle=True)
            df_out = self.perepoch(e,train_iter,test_iter,val_iter,train_data_shape,test_data_shape,val_data_shape)
            
            self.write_metadata(self.nndatapath + "/models/training_metadata_" + str(self.model_number) + ".hdf5","w",df_out)
        end = time.time()
        time_elapsed = (end-start)/60.
        torch.save(self.model.state_dict(), self.nndatapath + "/models/" + str(self.model_number) + ".pt")
        
        val_p, val_r, val_f, val_t, all_val_f = self.get_fscore(val_iter,val_data_shape)
        test_p, test_r, test_f, test_t, all_test_f = self.get_fscore(test_iter,test_data_shape)
                
        meta_handle = pandas_hdf5_handler(self.nndatapath + "/metadata.hdf5")
        trainmeta = meta_handle.read_df("train",read_metadata=True).metadata
        valmeta = meta_handle.read_df("val",read_metadata=True).metadata
        testmeta = meta_handle.read_df("test",read_metadata=True).metadata
        experiment_name = trainmeta["nndataset"]["experiment_name"]        
        train_dataname,train_org,train_micro,train_ttl_img = (trainmeta["nndataset"]["data_name"],trainmeta["global"]["Organism"],\
                                               trainmeta["global"]["Microscope"],trainmeta["nndataset"]["ttl_imgs"])
        val_dataname,val_org,val_micro,val_ttl_img = (valmeta["nndataset"]["data_name"],valmeta["global"]["Organism"],\
                                                      valmeta["global"]["Microscope"],valmeta["nndataset"]["ttl_imgs"])
        test_dataname,test_org,test_micro,test_ttl_img = (testmeta["nndataset"]["data_name"],testmeta["global"]["Organism"],\
                                                          testmeta["global"]["Microscope"],testmeta["nndataset"]["ttl_imgs"])
        
        train_loss,val_loss,test_loss = df_out['Train Loss'].tolist()[0],df_out['Val Loss'].tolist()[0],df_out['Test Loss'].tolist()[0]
        
        entry = [[experiment_name,self.model_number,train_dataname,train_org,train_micro,train_ttl_img,val_dataname,val_org,val_micro,val_ttl_img,\
                  test_dataname,test_org,test_micro,test_ttl_img,self.batch_size,self.layers,self.hidden_size,self.lr,self.momentum,\
                  self.weight_decay,self.dropout,self.w0,self.wm_sigma,train_loss,val_loss,val_p,val_r,val_f,val_t,all_val_f,test_loss,test_p,test_r,\
                  test_f,test_t,all_test_f,str(timestamp),self.numepochs,time_elapsed]]
                
        df_out = pd.DataFrame(data=entry,columns=['Experiment Name','Model #','Train Dataset','Train Organism','Train Microscope','Train # Images',\
                                                  'Val Dataset','Val Organism','Val Microscope','Val # Images',\
                                                  'Test Dataset','Test Organism','Test Microscope','Test # Images',\
                                                  'Batch Size','Layers','Hidden Size','Learning Rate','Momentum','Weight Decay',\
                                                  'Dropout',"W0 Weight","Wm Sigma",'Train Loss','Val Loss','Val Precision','Val Recall','Val F1 Score',\
                                                  'Val Threshold','Val F1 Cell Scores','Test Loss','Test Precision','Test Recall','Test F1 Score',\
                                                  'Test Threshold','Test F1 Cell Scores','Date/Time','# Epochs','Training Time (mins)'])
        
        df_out = df_out.set_index(['Experiment Name','Model #'], drop=True, append=False, inplace=False)
        df_out = df_out.sort_index()
        
        metalock = hdf5lock(self.nndatapath + "/model_metadata.hdf5",updateperiod=5.)
        metalock.lockedfn(self.write_metadata,"w",df_out)

class TrainingVisualizer:
    def __init__(self,trainpath,modeldbpath):
        self.trainpath = trainpath
        self.modelpath = trainpath + "/models"
        self.modeldfpath = trainpath + "/model_metadata.hdf5"
        self.modeldbpath = modeldbpath
        self.paramdbpath = modeldbpath+"/Parameters"
        self.update_dfs()
        if os.path.exists(self.modeldfpath):
            self.models_widget = qgrid.show_grid(self.model_df.sort_index())
        
    def update_dfs(self):
        df_idx_list = []
        for path in os.listdir(self.modelpath):
            if "training_metadata" in path:
                df_idx = int(path.split("_")[-1][:-5])
                df_idx_list.append(df_idx)
        df_list = []
        for df_idx in df_idx_list:
            dfpath = self.modelpath + "/training_metadata_" + str(df_idx) + ".hdf5"
            df_handle = pandas_hdf5_handler(dfpath)
            df = df_handle.read_df("data")
            df_list.append(copy.deepcopy(df))
            del df
        self.train_df = pd.concat(df_list)
        if os.path.exists(self.modeldfpath):
            modeldfhandle = pandas_hdf5_handler(self.modeldfpath)
            self.model_df = modeldfhandle.read_df("data").sort_index()
        
    def select_df_columns(self,selected_columns):
        df = copy.deepcopy(self.model_df)
        for column in df.columns.tolist():
            if column not in selected_columns:
                df = df.drop(column, 1)
        self.model_widget = qgrid.show_grid(df)

    def inter_df_columns(self):
        column_list = self.model_df.columns.tolist()
        inter = ipyw.interactive(self.select_df_columns,{"manual":True},selected_columns=ipyw.SelectMultiple(options=column_list,description='Columns to Display:',disabled=False))
        display(inter)

    def handle_filter_changed(self,event,widget):
        df = widget.get_changed_df().sort_index()
        
        all_model_indices = self.train_df.index.get_level_values("Model #").unique().tolist()
        current_model_indices = df.index.get_level_values("Model #").unique().tolist()
        
        all_epochs = []
        all_loss = []
        for model_idx in all_model_indices:
            if model_idx in current_model_indices:
                filter_df = df.loc[model_idx]
                epochs,loss = (filter_df.index.get_level_values("Epoch").tolist(),filter_df[self.losskey].tolist())
                all_epochs += epochs
                all_loss += loss
                self.line_dict[model_idx].set_data(epochs,loss)
                self.line_dict[model_idx].set_label(str(model_idx))
            else:
                epochs_empty,loss_empty = ([],[])
                self.line_dict[model_idx].set_data(epochs_empty,loss_empty)
                self.line_dict[model_idx].set_label('_nolegend_')
                
        self.ax.set_xlim(min(all_epochs), max(all_epochs)+1)
        self.ax.set_ylim(0, max(all_loss)*1.1)
        self.ax.legend()
        self.fig.canvas.draw()

    def inter_plot_loss(self,losskey):
        self.losskey = losskey
        self.fig,self.ax = plt.subplots()
        self.grid_widget = qgrid.show_grid(self.train_df.sort_index())
        current_df = self.grid_widget.get_changed_df()
        
        self.line_dict = {}
        for model_idx in current_df.index.get_level_values("Model #").unique().tolist():
            filter_df = current_df.loc[model_idx]
            epochs,loss = (filter_df.index.get_level_values("Epoch").tolist(),filter_df[losskey].tolist())
            line, = self.ax.plot(epochs,loss,label=str(model_idx))
            self.line_dict[model_idx] = line
            
        self.ax.set_xlabel("Epoch")
        self.ax.set_ylabel(losskey)
        self.ax.legend()
        
    def export_models(self):
        writedir(self.modeldbpath,overwrite=False)
        writedir(self.modeldbpath+"/Parameters",overwrite=False)
        modeldbhandle = pandas_hdf5_handler(self.modeldbpath + "/Models.hdf5")
        if "Models.hdf5" in os.listdir(self.modeldbpath):
            old_df = modeldbhandle.read_df("data")
            current_df = self.models_widget.get_changed_df()
            current_df = pd.concat([old_df, current_df])
        else:
            current_df = self.models_widget.get_changed_df()
        modeldbhandle.write_df("data",current_df)
        
        indices = current_df.index.tolist()
        exp_names = [str(item[0]) for item in indices]
        model_numbers = [str(item[1]) for item in indices]
        dates = [item.replace(" ","_") for item in current_df["Date/Time"].tolist()]
        
        for i in range(len(model_numbers)):
            exp_name,model_number,date = (exp_names[i],model_numbers[i],dates[i])
            shutil.copyfile(self.modelpath + "/" + str(model_number) + ".pt",self.paramdbpath+"/"+exp_name+"_"+model_number+"_"+date+".pt")

class UNet_Segmenter:
    def __init__(self,headpath,modeldbpath):#,min_obj_size=20):#,cell_threshold_scaling=1.,border_threshold_scaling=1.,\
             #    layers=3,hidden_size=64,batch_size=100,gpuon=False):
        
        self.headpath = headpath
        self.kymopath = headpath + "/kymograph"
        self.nnoutputpath = headpath + "/phasesegmentation"
        self.metapath = headpath + "/metadata.hdf5"
        self.meta_handle = pandas_hdf5_handler(self.metapath)
        globaldf = self.meta_handle.read_df("global",read_metadata=True)
        self.all_channels = globaldf.metadata['channels']
        
        self.modeldbpath = modeldbpath
        self.update_dfs()
        
        if os.path.exists(self.modeldbpath):
            self.models_widget = qgrid.show_grid(self.models_df.sort_index())
        
#         self.min_obj_size = min_obj_size
#         self.border_threshold_scaling = border_threshold_scaling
#         self.gpuon = gpuon
#         self.layers = layers
#         self.hidden_size = hidden_size
#         self.batch_size = batch_size
        
    def update_dfs(self):
        dfpath = self.modeldbpath + "/Models.hdf5"
        if os.path.exists(dfpath):
            df_handle = pandas_hdf5_handler(dfpath)
            self.models_df = df_handle.read_df("data")
            
    def select_df_columns(self,selected_columns):
        df = copy.deepcopy(self.models_df)
        for column in df.columns.tolist():
            if column not in selected_columns:
                df = df.drop(column, 1)
        self.model_widget = qgrid.show_grid(df)

    def inter_df_columns(self):
        column_list = self.models_df.columns.tolist()
        inter = ipyw.interactive(self.select_df_columns,{"manual":True},selected_columns=ipyw.SelectMultiple(options=column_list,description='Columns to Display:',disabled=False))
        display(inter)
        
    def select_model(self):
        current_df = self.model_widget.get_changed_df()
        print(current_df)
        
    def choose_seg_channel(self,seg_channel):
        self.seg_channel = seg_channel
                
#     def prepare_data(self,fov_num):
#         writedir(self.nnoutputpath,overwrite=False)
#         writepath = self.nnoutputpath + "/nninput_" + str(fov_num) + ".hdf5"
#         img_path = self.kymopath + "/kymo_" + str(fov_num) + ".hdf5"
#         with h5py.File(writepath,"w") as outfile:
#             with h5py.File(img_path,"r") as infile:
#                 keys = list(infile.keys())
#                 ex_data = infile[keys[0]+"/"+self.seg_channel]
#                 out_shape = (len(keys)*ex_data.shape[2],1,ex_data.shape[0],ex_data.shape[1])
#                 chunk_shape = (1,1,out_shape[2],out_shape[3])
#                 img_handle = outfile.create_dataset("img",out_shape,chunks=chunk_shape,dtype=float)
                
#                 for i,trenchid in enumerate(keys):
#                     img_arr = infile[trenchid+"/"+self.seg_channel][:]
#                     img_arr = np.moveaxis(img_arr,(0,1,2),(1,2,0))
#                     img_arr = np.expand_dims(img_arr,1)
#                     img_arr = img_arr.astype(float)
                    
#                     img_handle[i*ex_data.shape[2]:(i+1)*ex_data.shape[2]] = img_arr
    
#     def segment(self,fov_num):
#         torch.cuda.empty_cache()
#         self.model = UNet(1,3,layers=self.layers,hidden_size=self.hidden_size,withsoftmax=True)
        
#         if self.gpuon:
#             device = torch.device("cuda")
#             self.model.load_state_dict(torch.load(self.paramspath))
#             self.model.to(device)
#         else:
#             device = torch.device('cpu')
#             self.model.load_state_dict(torch.load(self.paramspath, map_location=device))
#         self.model.eval()
        
#         inputpath = self.nnoutputpath + "/nninput_" + str(fov_num) + ".hdf5"
#         outputpath = self.nnoutputpath + "/nnoutput_" + str(fov_num) + ".hdf5"
#         with h5py.File(inputpath,"r") as infile:
#             out_shape = tuple((infile["img"].shape[0],3,infile["img"].shape[2],infile["img"].shape[3]))
#             chunk_shape = infile["img"].chunks
#         data = SegmentationDataset(inputpath,training=False)
#         data_iter = DataLoader(data,batch_size=self.batch_size,shuffle=False) #careful
#         with h5py.File(outputpath,"w") as outfile:
#             img_handle = outfile.create_dataset("img",out_shape,chunks=chunk_shape,dtype=float)
#             print(len(data_iter))
#             for i,b in enumerate(data_iter):
#                 print(i)
#                 x = torch.Tensor(b['img'].numpy())
#                 if self.gpuon:
#                     x = x.cuda()
#                 fx = self.model.forward(x) #N,3,y,x
#                 img_handle[i*self.batch_size:(i+1)*self.batch_size] = fx.cpu().data.numpy()
        
#     def postprocess(self,fov_num):
#         nninputpath = self.nnoutputpath + "/nninput_" + str(fov_num) + ".hdf5"
#         nnoutputpath = self.nnoutputpath + "/nnoutput_" + str(fov_num) + ".hdf5"
#         segpath = self.nnoutputpath + "/seg_" + str(fov_num) + ".hdf5"
#         kymopath = self.kymopath + "/kymo_" + str(fov_num) + ".hdf5"
#         with h5py.File(kymopath,"r") as kymofile:
#             trench_num = len(kymofile.keys())
#             trenchids = list(kymofile.keys())
#         with h5py.File(segpath,"w") as outfile:
#             with h5py.File(nnoutputpath,"r") as infile:
#                 num_img = infile["img"].shape[0]
#                 y_shape,x_shape = (infile["img"].shape[2],infile["img"].shape[3])
#                 timepoints = int(num_img/trench_num)
#                 for trench in range(trench_num):
#                     print(trench)
#                     trenchid = trenchids[trench]
#                     trench_data = infile["img"][trench*timepoints:(trench+1)*timepoints] #t,3,y,x
#                     trench_output = []
#                     for t in range(timepoints):
#                         cell_otsu = sk.filters.threshold_otsu(trench_data[t,1])*self.cell_threshold_scaling
#                         border_otsu = sk.filters.threshold_otsu(trench_data[t,2])*self.border_threshold_scaling
#                         cell_mask = trench_data[t,1]>cell_otsu
#                         border_mask = trench_data[t,2]>border_otsu
#                         trench_arr = cell_mask*(~border_mask)
#                         del cell_mask
#                         del border_mask
#                         trench_arr = sk.morphology.remove_small_objects(trench_arr,min_size=self.min_obj_size)                        
#                         trench_output.append(trench_arr)
#                         del trench_arr
#                     trench_output = np.array(trench_output)
#                     trench_output = np.moveaxis(trench_output,(0,1,2),(2,0,1))
#                     outdset = outfile.create_dataset(trenchid, data=trench_output, chunks=(y_shape,x_shape,1), dtype=bool)
# #         os.remove(nninputpath)
# #         os.remove(nnoutputpath)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='TrenchRipper',
    version='0.0.3',
    author="Daniel Eaton",
    packages=setuptools.find_packages(),
    install_requires=[
        'pandas==1.0.3',
        'numpy>=1.18.1',
        'h5py>=1.10.4',
        'scipy>=1.4.1',
        'scikit-image>=0.16.2',
        'jupyterlab==2.0.1',
        'matplotlib>=3.1.3',
        'dask==2.12.0',
        'distributed==2.12.0',
        'dask-jobqueue==0.7.0',
        'tifffile>=2020.2.16',
        'ipywidgets==7.5.1',
        'pulp>=1.6.8',
        'fastparquet>=0.3.3',
        'torch>=1.4.0',
        'torchvision>=0.5.0',
        'cudatoolkit==10.0.130',
        'tables>=3.6.1',
        'scikit-learn>=0.22.1',
        'seaborn==0.10.0',
        'bokeh==1.4.0',
        'h5py_cache==1.0',
        'nd2reader==3.2.3',
        'parse==1.15.0',
        'qgrid==1.3.1',
        'opencv-python'
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DanielScottEaton/TrenchRipper",
    python_requires='>=3.7',
)